
package cyclicBuffer;
/**
 * Ein zyklischer Buffer fr int-Werte.
 * Durch den Konstrukturaufruf wird ein Buffer mit der angegebenen Gre
 * erzeugt. ber die Methode {@link #insert(int[])} knnen int-Werte
 * eingefgt werden. Eingefgt wird in die nchste als frei markierte Stelle im
 * Buffer. Da dies ein zyklischer Buffer ist, werden die Werte bei berschreitung
 * der Buffergre wieder vom Anfang beginnend eingefgt. Dadurch werden die ltesten
 * Werte berschrieben.
 * Beim Auslesen des Buffers werden die Werte in der Reihenfolge, wie sie eingefgt wurden
 * zurckgegeben.
 * <p/> 
 * Created on 12.11.2003
 * @author <A HREF="http://www.graviton.de">M. Serhat Cinar</A>
 * @version 11.06.2004
 */
public class CyclicIntBuffer {
	private int [] mValues;
	private int mActualInPos = 0;
	private int mInsertedValues = 0;
	
	/**
	 * Erzeugt einen zyklischen Buffer mit der gegebenen Anzahl an
	 * Speicherpltzen fr int-Werte.
	 * 
	 * @param size Anzahl der Speicherpltze fr int-Werte in diesem Buffer.
	 */
	public CyclicIntBuffer(int size){
		mValues = new int[size];
		init();
	}
	
	public void init(){
		for (int i=0; i<mValues.length; i++)
			mValues[i] = 0;
		mInsertedValues = 0;
		mActualInPos = 0;
	}
	
	public int getAmountInsertedValues(){
		return mInsertedValues;
	}
	
	public void insert(int pNewValue){
		mValues[mActualInPos] = pNewValue;
		mActualInPos = (mActualInPos+1) % mValues.length;
		mInsertedValues++;
	}
	
	/**
	 * Fgt neue Werte in den Buffer ein.
	 * Konkret werden die Positionen fr die neuen Werte wie folgt ermittelt:<br/>
	 * <pre>
	 * for (int i=0; i<newvalues.length; i++){
	 * 		mValues[mActualInPos] = newvalues[i];
	 * 		mActualInPos = (mActualInPos+1) % mValues.length;
	 * }
	 * </pre>
	 * 
	 * @param pNewValues Array mit int-Werten, die in der gegebenen 
	 * 			Reihenfolge eingefgt werden sollen.
	 */
	public void insert(int [] pNewValues){
		for (int i=0; i<pNewValues.length; i++){
			mValues[mActualInPos] = pNewValues[i];
			mActualInPos = (mActualInPos+1) % mValues.length;
			mInsertedValues++;
		}
		//Log.getInstance().writeln(this.getClass().getName()+".insert("+Formats.format(newvalues)+"): "+toString());
	}
	
	/**
	 * Liefert den aktuellen Inhalt des Buffers zurck.
	 * Konkret werden die Werte in der Reihenfolge ihres einfgens wiedergegeben.<br/>
	 * <pre>
	 * int [] outbuf = new int[mValues.length];
	 * for (int i=0; i<mValues.length; i++)
	 * 	outbuf[i] = mValues[(mActualInPos+i) % mValues.length];
	 * return outbuf;
	 * </pre>
	 * 
	 * @return Der aktuelle Inhalt des Buffers zyklisch ab der aktuellen Position.
	 */
	public int [] getBuffer(){
		int [] lOutBuffer = new int[mValues.length];
		for (int i=0; i<mValues.length; i++)
			lOutBuffer[i] = mValues[(mActualInPos+i) % mValues.length];
		return lOutBuffer;
	}
	
/*	public Object [] getBuffer(Object [] pObjectArray){
		pObjectArray = new Object[mValues.length];
		for (int i=0; i<mValues.length; i++)
			pObjectArray[i] = mValues[(mActualInPos+i) % mValues.length];
		return pObjectArray;
	}
	*/
	/**
	 * Liefert die Anzahl der Speicherpltze fr int-Werte.
	 * 
	 * @return Anzahl der Speicherpltze fr int-Werte.
	 */
	public int getSize(){	return mValues.length;}
	
	
	/**
	 * Erzeugt einen String mit allen int-Werten entsprechend der Methode getBuffer().
	 * 
	 * @see java.lang.Object#toString()
	 * @see #getBuffer()
	 */
	public String toString(){
		StringBuffer lSb = new StringBuffer();
		int [] lBuffer = getBuffer();
		for (int i=0; i<lBuffer.length; i++)
			lSb.append(lBuffer[i]+" ");
		return lSb.toString();
	}
}
