package logger;

import java.io.*;
import java.text.SimpleDateFormat;
import java.util.*;

/**
 * Klasse zum Erzeugen von Statusmeldungen
 * auf der Konsole und in eine Datei.<br/>
 * Es ist als Singleton-Pattern realisiert und wird im allgemeinen wie folgt verwendet:<br/>
 * <code>Log.getInstance().writeln(Eine Statusmeldung);</code><br/>
 * Optional kann fr jede Meldung auch automatisch die aufrufende Methode und Klasse
 * mit ausgegeben werden:<br/>
 * <code>Log.getInstance().writeln(Eine Statusmeldung, true);</code><br/>
 * ber die Propertiesdatei logger.properties knnen folgende Schlssel definiert werden:
 * <dir>
 *  <li><code>logtocli=true</code><br/>
 * 		Legt fest, ob die Meldungen in die Standardausgabe ausgegeben werden sollen. 
 * 		Mgliche Parameter sind true (Ausgabe aktivieren) oder false (keine Ausgabe machen).
 *  </li>
 *  <li><code>logtofile=true</code><br/>
 * 		Legt fest, ob die Meldungen in die Logdatei ausgegeben werden sollen. 
 * 		Mgliche Parameter sind true (Ausgabe aktivieren)oder false (keine Ausgabe machen).
 *  </li>
 *  <li><code>logdateformat=yyyy.MM.dd  HH:mm:ss</code><br/>
 * 		Legt das Format fr das Datum fest, das als erste Zeile in die Ausgaben
 * 		ausgegeben wird. Das Format muss der Klasse java.text.SimpleDateFormat entsprechen.
 *  </li>
 *  <li><code>fileprefix=log </code><br/>
 * 		Der Dateinamenprefix fr die Logdatei. 
 * 		Der Dateiname setzt sich aus dem Prefix, dem unter 
 * 		filenamedateformat definierten Datum und der Dateiendung zusammem.
 *  </li>
 *  <li><code>fileextension=.txt</code><br/>
 * 		Dateiendung fr die Logdatei.
 * 		Der Dateiname setzt sich aus dem Prefix, dem unter 
 * 		filenamedateformat definierten Datum und der Dateiendung zusammem.
 *  </li>
 *  <li><code>filepath=logs</code><br/>
 * 		Verzeichnis fr die Logdateien. Hier werden die Logdateien abgelegt.
 *  </li>
 *  <li><code>UseDateInFilename=true</code><br/>
 * 		Legt fest, ob das unter filenamedateformat angegebene Datum im Dateinamen
 * 		der Logdatei benutzt werden soll. 
 *  </li>
 *  <li><code>filenamedateformat=yyyy.MM.dd HH.mm.ss</code><br/>
 * 		Legt das Format fr das Datum fest, das im Namen der Logdatei benutzt werden soll.
 * 		Wird nur dann benutzt, wenn UseDateInFilename auf den Wert true gesetzt wurde.
 * 		Wichtig ist, dass Dateinamen dass Symbol ":" nicht enthalten drfen und daher auch
 * 		im Datum nicht stehen drfen.
 * 		Der Dateiname setzt sich aus dem Prefix, dem unter 
 * 		filenamedateformat definierten Datum und der Dateiendung zusammem.
 * 		Das Format muss der Klasse java.text.SimpleDateFormat entsprechen.
 *  </li>
 * </dir>
 * <p/>
 * Created on 01.04.2004
 * @author <A HREF="http://www.graviton.de">M. Serhat Cinar</A>
 * @version 13.08.2004
 * @see java.text.SimpleDateFormat
 */
public class Log {
	private static Log mInstance = null;
	
	private static ResourceBundle mResourceBundle = ResourceBundle.getBundle("logger");
	private static SimpleDateFormat smLogSimpleDateFormat;
	private static SimpleDateFormat smFileNameSimpleDateFormat;
	static{
		smLogSimpleDateFormat = new SimpleDateFormat(mResourceBundle.getString("logdateformat"));
		smFileNameSimpleDateFormat = new SimpleDateFormat(mResourceBundle.getString("filenamedateformat"));
	}

	private BufferedWriter mLogFile = null;
	private boolean mLogToFile = true;
	private boolean mLogToCLI = true;
	private String mLogFileName;
	private String mLogFullPath;
	
	private Log() {
		if (mResourceBundle.getString("logtocli").equalsIgnoreCase("true"))
			mLogToCLI = true;
		else mLogToCLI = false;
		if (mResourceBundle.getString("logtofile").equalsIgnoreCase("true"))
			mLogToFile = true;
		else mLogToFile = false;

		boolean lDateInFilename = true;
		if (mResourceBundle.getString("UseDateInFilename").equalsIgnoreCase("true"))
			lDateInFilename = true;
		if (lDateInFilename)
			mLogFileName = mResourceBundle.getString("fileprefix")+
				smFileNameSimpleDateFormat.format(Calendar.getInstance().getTime())+
				mResourceBundle.getString("fileextension");
		else mLogFileName = mResourceBundle.getString("fileprefix")+mResourceBundle.getString("fileextension");
		mLogFullPath = mResourceBundle.getString("filepath")+"/"+mLogFileName; 
		System.out.println(this.getClass().getName()+" CLI: "+mLogToCLI+" FILE: "+mLogToFile+" ["+mLogFullPath+"]");
	}
	
	/**
	 * Liefert die aktuelle Instanz des Loggers zurck.
	 * 
	 * @return Die aktuelle Instanz des Loggers.
	 */
	public static Log getInstance() {
		if (mInstance == null) {
			try {
				mInstance = new Log();
				if (mInstance.mLogToFile && mInstance.mLogFile == null)
					mInstance.mLogFile = new BufferedWriter(new FileWriter(mInstance.mLogFullPath, true));
				
				// start logsession
				mInstance.writeln("");
				mInstance.writeln("[ open new session " + now() + " ]");
				mInstance.writeln("");
			}
			catch (IOException e) {
				System.err.println("Couldn't open logfile " + mInstance.mLogFileName + " to write to: "+e.toString());
				mInstance.mLogFile = null;
				mInstance.mLogToFile = false;
			}
			catch (java.security.AccessControlException e) {
				System.err.println("Couldn't open logfile " + mInstance.mLogFileName + " to write to: "+e.toString());
				mInstance.mLogFile = null;
				mInstance.mLogToFile = false;
			}
		}
		return mInstance;
	}
	
	public BufferedWriter getFileWriter() {
		return mLogFile;
	}
	
	/**
	 * Gibt einen String ohne Zeilenvorschub aus. 
	 * 
	 * @param s String, der auszugeben ist.
	 */
	public synchronized void write(String s) {
		if ((mLogFile != null) && mLogToFile) {
			try {
				mLogFile.write(s);
				mLogFile.flush();
			}
			catch (IOException e) {
				System.err.println(
					"Error while writing to logfile " + mLogFileName+": "+e.toString());
				mLogFile = null;
				mLogToFile = false;
			}
		}
		if (mLogToCLI) System.out.print(s);
	}

	/**
	 * Gibt einen String mit Zeilenvorschub aus. 
	 * 
	 * @param s String, der auszugeben ist.
	 */
	public synchronized void writeln(String s) {
		write(s + "\n");
	}
	
	
	/**
	 * Gibt einen String ohne Zeilenvorschub aus und fgt
	 * vor dem String den vollen Namen der Klasse und der Methode ein,
	 * die diese Ausgabe veranlasst hat.<br/>
	 * Die Ausgabe sieht dann etwa so aus:<br/>
	 * myPackage.myClass.someMethod:thestring
	 * 
	 * @param s String, der auszugeben ist.
	 * @param debugCaller
	 */
	public synchronized void write(String s, boolean debugCaller){
		if (debugCaller){
			Throwable t = new Throwable();
			StackTraceElement elements[] = t.getStackTrace();
			String lClass = elements[1].getClassName();
			String lMethod = elements[1].getMethodName();
			write(lClass+"."+lMethod+":"+s);
		}
		else write(s);
	}
	
	/**
	 * Gibt einen String mit Zeilenvorschub aus und fgt
	 * vor dem String den vollen Namen der Klasse und der Methode ein,
	 * die diese Ausgabe veranlasst hat.<br/>
	 * Die Ausgabe sieht dann etwa so aus:<br/>
	 * myPackage.myClass.someMethod:thestring
	 * 
	 * @param s String, der auszugeben ist.
	 * @param debugCaller
	 */
	public synchronized void writeln(String s, boolean debugCaller){
		if (debugCaller){
			Throwable t = new Throwable();
			StackTraceElement elements[] = t.getStackTrace();
			String lClass = elements[1].getClassName();
			String lMethod = elements[1].getMethodName();
			writeln(lClass+"."+lMethod+":"+s);
		}
		else writeln(s);
	}
	
	public synchronized void write(String s, int pCallerLevel){
		Throwable t = new Throwable();
		StackTraceElement elements[] = t.getStackTrace();
		if (pCallerLevel > elements.length) pCallerLevel = elements.length;
		String lClass;
		String lMethod;
		for (int i=pCallerLevel; i>=1; i--){
			lClass = elements[i].getClassName();
			lMethod = elements[i].getMethodName();
			write(lClass+"."+lMethod+"()->");
			if (i==1) write(lClass+"."+lMethod+":"+s);
		}
	}
	
	public synchronized void writeln(String s, int pCallerLevel){
		Throwable t = new Throwable();
		StackTraceElement elements[] = t.getStackTrace();
		if (pCallerLevel > elements.length) pCallerLevel = elements.length;
		String lClass;
		String lMethod;
		for (int i=pCallerLevel; i>=1; i--){
			lClass = elements[i].getClassName();
			lMethod = elements[i].getMethodName();
			write(lClass+"."+lMethod+"()->");
			if (i==1) writeln(lClass+"."+lMethod+":"+s);
		}
	}
	
	/**
	 * Liefert das aktuelle Datum und die Uhrzeit als String zurck, in dem Format,
	 * wie es im Schlssel logdateformat in der Propertiesdatei festgelegt wurde.
	 * 
	 * @return Das aktuelle Datum als String.
	 */
	public static synchronized String now() {
		return smLogSimpleDateFormat.format(Calendar.getInstance().getTime());
	}
}
