package mscJNeuralNet.connectors;

import java.util.ResourceBundle;

import observerPattern.Observable;

import formats.Formats;

import logger.Log;
import mscJNeuralNet.net.Net;

/**
 * Kantengewichtsinitialisierung mit Zufallswerten entsprechend den eingehenden Verbindungen
 * eines Neurons.
 * <p/>
 * Aus <i>Prof. Dr. H. Westenberger, Skript Neuronale Netze 99V011</i>:
 * <p/> 
 * <h3>Kapitel: 2.5.1 Symmetry breaking</h3><br/>
 * "Fr Feedforward-Netze, die ebenenweise aufgebaut sind, sollten die Gewichte
 * einer Verbindungsschicht beim Start nicht smtlich gleich initialisiert werden.
 * In einem ff1-Netz mit einem Hidden-Layer wrden nmlich aus
 * Symmetriegrnden die Kantengewichte, die von einem Input Knoten ausgehen
 * ebenso wie alle Kanten zu einem Outputknoten gleich bleiben
 * Lsung: Bei der Initialiserung der Gewichte verwende man Zufallszahlen.
 * Optimal sind die Kantengewichte, wenn die Betragssumme der
 * Kantengewichte, die zu einem Neuron fhren, betragsmig klein bleiben,
 * allenfalls in der Grenordnung von 1 liegen. Damit wird der Netzinput in der
 * Nhe von 0 liegen, und die sigmoide Transferfunktion relativ groe
 * Ableitungen erreichen. In dieser Situation kann sich das Netz schnellstens
 * anpassen. Aus diesem Grund empfiehlt sich, die Zufallszahlen fr die
 * Kantenvorbelegung zu einem Neuron mit N Vorgngerneuronen aus dem
 * Intervall [-1/N; 1/N], zu whlen."
 * <p/>
 * Im folgenden Beispiel wird der RandomSymmetryBreakingNetConnector benutzt, 
 * um die Kantengewichte mit Zufallswerten entsprechend zu initialisieren:<br/>
 * <pre>
 * int [] lLayerSizes = {2, 1}; 
 * Net myNet = new Net();
 * INetConnector lRndSymBreakCon = new RandomSymmetryBreakingNetConnector();
 * lRndSymBreakCon.connectNet(myNet, null); 
 * <p/>
 * Created on 02.06.2004
 * @author <A HREF="http://www.graviton.de">M. Serhat Cinar</A>
 * @version 02.06.2004
 * @see INetConnector
 */
public class RandomSymmetryBreakingNetConnector implements INetConnector {
	private static final ResourceBundle smResourceBundle = ResourceBundle.getBundle("debug");
	private static final boolean smLogMethodCall;
	static{
		if (smResourceBundle.getString("showMethodCall.NetConnector").equalsIgnoreCase("true"))
			smLogMethodCall = true;
		else smLogMethodCall = false;
	}
	
	/**
	 * Gegebene Parameter werden nicht bercksichtigt. Daher sollte der 
	 * Parameterarray als null bergeben werden.<br/>
	 * Nach der Initialisierung werden die Observer des Netzes benachrichtigt.
	 * 
	 * @see mscJNeuralNet.connectors.INetConnector#connectNet(Net, double[])
	 * @see observerPattern.ObserverManager#notifyObservers(Observable)
	 */
	public void connectNet(Net pNet, double [] pParameters) {
		double[][] lWeights = pNet.getWeights();
		int [][] lLayerNeurons = pNet.getLayerNeurons();
		if (smLogMethodCall) Log.getInstance().writeln(this.getClass().getName()+".createConnections("+Formats.format(pParameters)+")");
		
		// weightmatrix entsprechend mLayerNeurons erzeugen
		
		int lInputtingLayer, lOutputtingLayer;
		int lInputtingNeuronIndex, lOutputtingNeuronIndex;
		
		// first connect BIAS to all but inputlayer
		for (lInputtingLayer = 1; lInputtingLayer<lLayerNeurons.length; lInputtingLayer++){
			for (lInputtingNeuronIndex=0; lInputtingNeuronIndex<lLayerNeurons[lInputtingLayer].length; lInputtingNeuronIndex++){
				if (lInputtingLayer==1)
					lWeights[0][lLayerNeurons[lInputtingLayer][lInputtingNeuronIndex]] = 
						Net.getRandomWeightValue(1D/(lLayerNeurons[lInputtingLayer-1].length));
				else
					lWeights[0][lLayerNeurons[lInputtingLayer][lInputtingNeuronIndex]] = 
						Net.getRandomWeightValue(1D/(lLayerNeurons[lInputtingLayer-1].length+1));
					//																		+1 wegen BIAS
			}
		}

		for (lOutputtingLayer=0; lOutputtingLayer<lLayerNeurons.length-1; lOutputtingLayer++){
			lInputtingLayer = lOutputtingLayer+1;
			for (lOutputtingNeuronIndex=0; lOutputtingNeuronIndex<lLayerNeurons[lOutputtingLayer].length; lOutputtingNeuronIndex++){
				for (lInputtingNeuronIndex=0; lInputtingNeuronIndex<lLayerNeurons[lInputtingLayer].length; lInputtingNeuronIndex++){
					if (lInputtingLayer==1)
						lWeights
							[lLayerNeurons[lOutputtingLayer][lOutputtingNeuronIndex]]
							[lLayerNeurons[lInputtingLayer][lInputtingNeuronIndex]] = 
								Net.getRandomWeightValue(1D/(lLayerNeurons[lOutputtingLayer].length));
					else
						lWeights
							[lLayerNeurons[lOutputtingLayer][lOutputtingNeuronIndex]]
							[lLayerNeurons[lInputtingLayer][lInputtingNeuronIndex]] = 
								Net.getRandomWeightValue(1D/(lLayerNeurons[lOutputtingLayer].length+1));
				}
			}
		}
		if (smResourceBundle.getString("showCreateConnections").equalsIgnoreCase("true"))
			Log.getInstance().writeln(Formats.format(lWeights));
		
		pNet.getObserverManager().notifyObservers(pNet);
	}
	
	public String toString(){	return "Random Symmetrybreaking Connector";}
}
