package mscJNeuralNet.gui.subComponents;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Rectangle;
import java.util.Vector;

import javax.swing.JPanel;

import observerPattern.Observable;
import observerPattern.ObserverManager;

/**
 * Ein Panel zur Darstellung eines Graphen aus einzelnen Hhenangaben.
 * Diese Komponente wird von der Komponente {@link mscJNeuralNet.gui.JNetErrorGraphView}
 * benutzt.
 * <p/>
 * <img src="../../../images/jgraphview.gif"/><p/>
 * Created on 10.06.2004
 * @author <A HREF="http://www.graviton.de">M. Serhat Cinar</A>
 * @version 19.08.2004
 * @see mscJNeuralNet.gui.JNetErrorGraphView
 */
public class JGraphView extends JPanel implements Observable{
	private ObserverManager mObserverManager = new ObserverManager();
	
	private Vector mHeights;
	// ein pixel entspricht mUnitPerHeightPixel Einheiten auf der Y-Achse
	private double mUnitPerHeightPixel;
	// horizontale grid linie jede mHeightGridsEvery Einheiten
	private double mHeightGridsEvery;
	private double mHeightGridsPixelStep;
	private double mUnitPerWidthPixel;
	// gridline every 50 horizontal unit
	private double mWidthGridsEvery;
	private double mWidthGridsPixelStep;
	
	/**
	 * Erzeugt ein Panel zur Darstellung eines Graphen.
	 * 
	 * @param pUnitPerWidthPixel Einheiten pro Pixel Breite.
	 * @param pUnitPerHeightPixel Einheiten pro Pixel Hhe.
	 * @param pWidthGridsEvery Horizontale Raster alle x Einheiten.
	 * @param pHeightGridsEvery Vertikale Raster alle x Einheiten.
	 */
	public JGraphView(double pUnitPerWidthPixel, double pUnitPerHeightPixel, double pWidthGridsEvery, double pHeightGridsEvery){
		super();
		setBackground(Color.WHITE);
		mHeights = new Vector();
		mUnitPerWidthPixel = pUnitPerWidthPixel;
		mUnitPerHeightPixel = pUnitPerHeightPixel;
		mWidthGridsEvery = pWidthGridsEvery;
		mHeightGridsEvery = pHeightGridsEvery;
		mHeightGridsPixelStep = (1/mUnitPerHeightPixel)*mHeightGridsEvery;
		mWidthGridsPixelStep = (1/mUnitPerWidthPixel)*mWidthGridsEvery;
		
// 		scrollRectToVisible(new Rectangle(0, 0, getPreferredSize().width, getPreferredSize().height));
	}
	
	/**
	 * Fgt einen weiteren Punkt zum Graphen hinzu. Der Punkt wird als 
	 * Integer-Objekt bergeben, der die y-Koordinate (in Pixeln) darstellt.
	 * Die x-Koordinate wird ber die Position / Reihenfolge, in welcher der
	 * Punkt hinzugefgt wurde, bestimmt.
	 *  
	 * @param pInt Integer-Objekt, welches die Hhe des nchsten Punktes bestimmt.
	 */
	public void addData(Integer pInt){
		mHeights.add(pInt);
		repaint();
	}
	
	/**
	 * Fgt die Punkte aus dem Vector in der gegebenen Reihenfolge
	 * dem Graphen hinzu. 
	 * 
	 * @param pData A Vector that contains Integer objects.
	 */
	public void addData(Vector pData){
		for (int i=0; i<pData.size(); i++)
			mHeights.add(pData.get(i));
		repaint();
	}
	
	/**
	 * Legt die Skalierung der x-Achse zu den Einheiten fest.
	 * Jeder Pixel in der x-Achse reprsentiert die angegebene
	 * Anzahl an Einheiten.
	 * 
	 * @param pUnitPerWidthPixel Anzahl Einheiten pro Pixel in der x-Achse. 
	 */
	public void setWidthScale(double pUnitPerWidthPixel){
		mUnitPerWidthPixel = pUnitPerWidthPixel;
		mWidthGridsPixelStep = (1/mUnitPerWidthPixel)*mWidthGridsEvery;
	}
	
	/**
	 * Liefert die Skalierung der x-Achse zu den Einheiten zurck.
	 * Jeder Pixel in der x-Achse reprsentiert die zurckgegebene
	 * Anzahl an Einheiten.
	 * 
	 * @return Anzahl Einheiten pro Pixel in der x-Achse. 
	 */
	public double getWidthScale(){
		return mUnitPerWidthPixel;
	}
	
	/**
	 * Liefert die Anzahl der Pixel zwischen zwei horizontalen Rasterlinien.
	 * Intern wird der Wert als (1/getWidthScale())*getWidthGridsEvery() berechnet.
	 * 
	 * @return Die Anzahl der Pixel zwischen zwei horizontalen Rasterlinien.
	 */
	public double getWidthPixelStep(){
		return mWidthGridsPixelStep;
	}
	
	/**
	 * Legt die Anzahl der Einheiten fest, bei denen ein horizontaler Raster gezeichnet werden soll.
	 * 
	 * @param pWidthGridsEvery Anzahl der Einheiten, bei denen ein horizontaler Raster gezeichnet werden soll.
	 */
	public void setWidthGridsEvery(double pWidthGridsEvery ){
		mWidthGridsEvery = pWidthGridsEvery;
		mWidthGridsPixelStep = (1/mUnitPerWidthPixel)*mWidthGridsEvery;
	}
	
	/**
	 * Liefert die Anzahl der Einheiten, bei denen ein horizontaler Raster gezeichnet werden soll.
	 * 
	 * @return Die Anzahl der Einheiten, bei denen ein horizontaler Raster gezeichnet werden soll.
	 */
	public double getWidthGridsEvery(){
		return mWidthGridsEvery;
	}
	
	/**
	 * Legt die Skalierung der y-Achse zu den Einheiten fest.
	 * Jeder Pixel in der y-Achse reprsentiert die angegebene
	 * Anzahl an Einheiten.
	 * 
	 * @param pUnitPerHeightPixel Anzahl Einheiten pro Pixel in der y-Achse. 
	 */
	public void setHeightScale(double pUnitPerHeightPixel){
		mUnitPerHeightPixel = pUnitPerHeightPixel;
		mHeightGridsPixelStep = (1/mUnitPerHeightPixel)*mHeightGridsEvery;
	}
	
	/**
	 * Liefert die Skalierung der y-Achse zu den Einheiten zurck.
	 * Jeder Pixel in der y-Achse reprsentiert die zurckgegebene
	 * Anzahl an Einheiten.
	 * 
	 * @return Anzahl Einheiten pro Pixel in der y-Achse. 
	 */
	public double getHeightScale(){
		return mUnitPerHeightPixel;
	}
	
	/**
	 * Liefert die Anzahl der Pixel zwischen zwei vertikalen Rasterlinien.
	 * Intern wird der Wert als (1/getHeightScale())*getHeightGridsEvery() berechnet.
	 * 
	 * @return Die Anzahl der Pixel zwischen zwei vertikalen Rasterlinien.
	 */
	public double getHeightPixelStep(){
		return mHeightGridsPixelStep;
	}
	
	/**
	 * Legt die Anzahl der Einheiten fest, bei denen ein vertikaler Raster gezeichnet werden soll.
	 * 
	 * @param pHeightGridsEvery Anzahl der Einheiten, bei denen ein vertiakler Raster gezeichnet werden soll.
	 */
	public void setHeightGridsEvery(double pHeightGridsEvery ){
		mHeightGridsEvery = pHeightGridsEvery;
		mHeightGridsPixelStep = (1/mUnitPerHeightPixel)*mHeightGridsEvery;
	}
		
	/**
	 * Liefert die Anzahl der Einheiten, bei denen ein vertikaler Raster gezeichnet werden soll.
	 * 
	 * @return Die Anzahl der Einheiten, bei denen ein vertikaler Raster gezeichnet werden soll.
	 */
	public double getHeightGridsEvery(){
		return mHeightGridsEvery;
	}
		
	/**
	 * Lscht alle Datenpunkte, so dass der Graph leer ist.
	 */
	public void reset(){
		mHeights = new Vector();
	}
	
	private int mDataMaxIndex;
	private Rectangle mDrawingArea;
	private int mInt;
	private double mScaleNum;
	private double mX;
	private double mY ;
	private int mValueHeightSource;
	private int mValueHeightTarget;
	private int mHeight;
	private int lDrawTo;
	
	/* (non-Javadoc)
	 * @see javax.swing.JComponent#paintComponent(java.awt.Graphics)
	 */
	public void paintComponent(Graphics pG){
		super.paintComponent(pG);
		mDrawingArea = pG.getClipBounds();
		mDataMaxIndex = mHeights.size()-1;
		
		// resize if panel is to small
		if (mDataMaxIndex > getPreferredSize().width){
			// create bigger image
			setPreferredSize(new Dimension(Math.max(getPreferredSize().width+200, mDataMaxIndex), getPreferredSize().height));
			revalidate();
			// TODO nur horizontal verschieben, nicht vertikal
			scrollRectToVisible(new Rectangle(getPreferredSize().width-mDrawingArea.width, mDrawingArea.y, getPreferredSize().width, mDrawingArea.height));
			mObserverManager.notifyObservers(this);
		}
		
		mScaleNum = 0;
		pG.setColor(Color.LIGHT_GRAY);
		
		// vertical grids
		for (mX = 0; mX < getPreferredSize().width; mX += mWidthGridsPixelStep){
			mInt = Math.round((float)mX);
			pG.drawLine(mInt, 0, mInt, getPreferredSize().height);
			mScaleNum += mWidthGridsEvery;
		}
		
		// horizontal grid
		mScaleNum = 0;
		for (mY = 0; mY < getPreferredSize().height; mY += mHeightGridsPixelStep){
			mInt = Math.round((float)mY);
			pG.drawLine(0, mInt, getPreferredSize().width, mInt);
			mScaleNum += mHeightGridsEvery;
		}
		
		// data
		mHeight = getPreferredSize().height;		
		pG.setColor(Color.BLACK);

/*
 
 		if (mDataMaxIndex > 0){
			mValueHeightSource = mHeight-((Integer)mHeights.get(0)).intValue();
			if (mDataMaxIndex > 1){
				// mehrere punkte gegeben
				for (mInt = 1; mInt < mDataMaxIndex; mInt++){
					mValueHeightTarget = mHeight-((Integer)mHeights.get(mInt)).intValue();
					pG.drawLine(mInt-1, mValueHeightSource, mInt, mValueHeightTarget);
					pG.drawLine(mInt-1, mValueHeightSource+1, mInt, mValueHeightTarget+1);
					mValueHeightSource = mValueHeightTarget; 
				}
			}
			else{
				// nur ein punkt gegeben
				pG.drawRect(0, mValueHeightSource, 1, mValueHeightSource-1);
			}
		}
*/

		if (mDrawingArea.x < mHeights.size()){
			mValueHeightSource = mHeight-((Integer)mHeights.get(mDrawingArea.x)).intValue();
			// mehrere punkte gegeben
			lDrawTo = Math.min(mDrawingArea.x+mDrawingArea.width, mDataMaxIndex);
			for (mInt = mDrawingArea.x+1; mInt <= lDrawTo; mInt++){
				mValueHeightTarget = mHeight-((Integer)mHeights.get(mInt)).intValue();
				pG.drawLine(mInt-1, mValueHeightSource, mInt, mValueHeightTarget);
				pG.drawLine(mInt-1, mValueHeightSource+1, mInt, mValueHeightTarget+1);
				mValueHeightSource = mValueHeightTarget; 
			}
		}
		else if (mDrawingArea.x == (mHeights.size()-1) && mHeights.size() > 0){
			mValueHeightSource = mHeight-((Integer)mHeights.get(mDrawingArea.x)).intValue();
		}
//		System.out.println(this.getClass().getName()+".paintComponent().FINISHED");
	}

	/* (non-Javadoc)
	 * @see observerPattern.Observable#getObserverManager()
	 */
	public ObserverManager getObserverManager() {
		return mObserverManager;
	}
}

