package mscJNeuralNet.net;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.Reader;
import java.io.StringReader;
import java.io.StringWriter;
import java.util.Random;
import java.util.ResourceBundle;

import formats.Formats;

import observerPattern.Observable;
import observerPattern.ObserverManager;

import mscJNeuralNet.trainingAlgorithms.*;

import logger.Log;

/**
 * Das eigentliche Multilayered Perceptron (MLP).
 * Diese Klasse enthlt die Kantengewichte des Netzes sowie die Logik zur
 * Berechnung der Netzausgabe zu eine Eingabe.
 * Weiterhin wird auch das eigentliche Training, die Reihenfolge der
 * Berechnung der Gewichte und der rckpropagierte Fehlergradient fr jedes Neuron hier ermittelt.<br/>
 * Zustzlich enthlt diese Klasse Methoden zum Laden und Speichern eines Netzes
 * sowie Hilfsmethoden wie die Transferfunktion und die Ableitung der Transferfunktion.
 * <p/>
 * Es wird immer ein Netz mit BIAS und Fullconnection erzeugt. Das BIAS wird als
 * ein zustzliches Neuron der Eingabeschicht mit dem konstanten Ausgabewert 1.0
 * realisiert, dass eine Verbindung zu allen Neuronen der verborgenen und der Ausgabeschicht besitzt.
 * <p/>
 * Ein Beispiel fr den Einsatz wird unter {@link mscJNeuralNet} gegeben.
 * <p/>
 * ber die Schnittstelle {@link observerPattern.Observable} knnen Observer nderungen
 * des Netzes berwachen.
 * <p/>
 * Created on 29.03.2004
 * @author <A HREF="http://www.graviton.de">M. Serhat Cinar</A>
 * @version 19.06.2004
 * @see mscJNeuralNet
 * @see observerPattern.Observable
 */
public class Net implements Observable{
	private static final ResourceBundle smMSCResourceBundle = ResourceBundle.getBundle("mscJNeuronalNet_Text");
	private static final Random smRnd = new Random(System.currentTimeMillis());
	
	private double [][] mWeights;
    public static final double BIAS_VALUE = 1D;
	private static ITransferfunction mTransfer = new TransferTanh();
    
	
	public static void setTransferfunction(ITransferfunction pTransfer) {
		mTransfer = pTransfer;
	}
	
    // netzeingabe redundant, da bereits in mNeuronInput enthalten
    // private double [] mNetInput;
    // netzausgabe redundant, da bereits in mNeuronOutput enthalten
	// private double [] mNetOutput;
	
    // gesamte anzahl neuronen inklusive BIAS
    private int mAmountNeurons;
	// input einzelner neuronen = net[]
	private double [] mNeuronInput; // INIT OK
	// output einzelner neuronen
	private double [] mNeuronOutput; // INIT OK

	// neuronen der layer [layer][neuron1-x] mit BIAS
	private int [][] mLayerNeurons; // INIT OK
	
	private ObserverManager mObserverManager = new ObserverManager();
	
	private static final ResourceBundle smResourceBundle = ResourceBundle.getBundle("debug");
	private static final boolean smLogMethodCall;
	
	public Net cloned(){
		try {
			StringWriter lSOut = new StringWriter();
			PrintWriter lOut = new PrintWriter(lSOut);
			save(this, lOut);
			String lS = lSOut.toString();
			return load(new StringReader(lS));
		}
		catch(IOException e) {return null;}
	}
	
	static{
		if (smResourceBundle.getString("showMethodCall.Net").equalsIgnoreCase("true"))
			smLogMethodCall = true;
		else smLogMethodCall = false;
	}
	
	/**
	 * Erzeugt ein Netz mit der angegebenen Layerstrucktur.
	 * Der Paramter-Array wird wie folgt interpretiert:<br/>
	 * Arrayindex 0: Gre der Eingabeschicht<br/>
	 * Arrayindex letzte Position: Gre der Ausgabeschicht<br/>
	 * Alle Arrayindizes dazwischen: Gre der verborgenen Schichten<br/>
	 * Es sind mindestens zwei Werte (Eingabe- / Ausgabeschicht) notwendig,
	 * um ein Netz zu erzeugen.
	 * 
	 * @param pLayerSizes Die Gren der einzelnen Schichten.
	 */
	public Net(int [] pLayerSizes){
		mAmountNeurons = 1; // anzahl neuronen mit BIAS
		for (int i=0; i<pLayerSizes.length; i++)
			mAmountNeurons += pLayerSizes[i];
		
		// create mWeights matrix and initialize with zero
		mWeights = new double[mAmountNeurons][mAmountNeurons];
		for (int lOutputingNeuronIndex = 0; lOutputingNeuronIndex<mAmountNeurons; lOutputingNeuronIndex++)
			for (int lInputingNeuronIndex = 0; lInputingNeuronIndex<mAmountNeurons; lInputingNeuronIndex++)
				mWeights[lOutputingNeuronIndex][lInputingNeuronIndex] = 0D;
		
		// create & init neurons inputvalues net[]
		// BIAS & inputneuronen haben keinen input.
		// sie bleiben stndig bei 0
		mNeuronInput = new double[mAmountNeurons];
		for (int i=0; i<mNeuronInput.length; i++)
			mNeuronInput[i] = 0D;
		
		// create & init neurons outputvalues out[]
		// BIAS hat konstanten outputwert
		// inputneuronen haben als output das eingabepattern
		mNeuronOutput = new double[mAmountNeurons];
		mNeuronOutput[0] = BIAS_VALUE;
		for (int i=1; i<mNeuronOutput.length; i++)
			mNeuronInput[i] = 0D;
		
		// neuronen der layer [layer][neuron1-x] mit BIAS
		// inputlayer mit BIAS
		mLayerNeurons = new int[pLayerSizes.length][];
		int lActualNeuronID = 0;
		mLayerNeurons[0] = new int[pLayerSizes[0]+1];
		for (int j=0; j<mLayerNeurons[0].length; j++)
			mLayerNeurons[0][j] = lActualNeuronID++;
		// restliche layer
		for (int i=1; i<pLayerSizes.length; i++){
			mLayerNeurons[i] = new int[pLayerSizes[i]];
			for (int j=0; j<mLayerNeurons[i].length; j++)
				mLayerNeurons[i][j] = lActualNeuronID++;
		}
	}
	
	/**
	 * Liefert die Ausgabe ALLER Neuronen zur gegebenen Eingabe.
	 * Die Ausgabe aller Neuronen wird zur Berechnung
	 * des Fehlergradienten bentigt.
	 * Hierzu wird intern die Methode {@link #getNetOutput(double[])}
	 * aufgerufen.
	 *  
	 * @param pInputPattern Eingabemsuter.
	 * @return Array mit der Ausgabe eines jeden Neurons.
	 * @throws PatternDoesNotMatchNetException Falls eines der gegebenen Lerndatenstze
	 * 		nicht der Eingabe-/ Ausgabeschichtgre des Netzes entspricht.
	 */
	public double [] getNeuronsOutput(double[] pInputPattern) throws PatternDoesNotMatchNetException{
		if (smLogMethodCall) Log.getInstance().writeln(this.getClass().getName()+".getNeuronsOutput("+Formats.format(pInputPattern)+")");
		getNetOutput(pInputPattern);
		return mNeuronOutput;
	}
	
	/**
	 * Berechnet die Ausgabe des Netzes zur gegebenen Eingabe (Ist-Werte).
	 * Das Eingabemuster wird vorwrts durch das Netz propagiert,
	 * um die Ausgabe zu berechnen.
	 * 
	 * @param pInputPattern Eingabemuster.
	 * @return Die Netzausgabe zum gegebenen Eingabemuster.
	 * @throws PatternDoesNotMatchNetException Falls eines der gegebenen Lerndatenstze
	 * 		nicht der Eingabe-/ Ausgabeschichtgre des Netzes entspricht.
	 */
	public double [] getNetOutput(double[] pInputPattern) throws PatternDoesNotMatchNetException{
//		if (smLogMethodCall) Log.getInstance().writeln(this.getClass().getName()+".getNetOutput("+Formats.format(pInputPattern)+")");
//		boolean lLog;
/*		if (smResourceBundle.getString("showGetNetOutput").equalsIgnoreCase("true"))
			lLog = true;
		else lLog = false;
*/				
		if (pInputPattern.length != getInputLayerSize())
			throw new PatternDoesNotMatchNetException(
					smMSCResourceBundle.getString("INPUTPATTERN_LENGTH")+": "+pInputPattern.length+" "+
					smMSCResourceBundle.getString("INPUTLAYER_LENGTH")+": "+getInputLayerSize()
				);
		double lOutput [] = new double[mLayerNeurons[mLayerNeurons.length-1].length];
		// input anlegen (BIAS bereits gesetzt)
		for (int i=1; i<mLayerNeurons[0].length; i++)
			mNeuronOutput[i] = pInputPattern[i-1];
		
		int lOutputtingLayer, lInputtingLayer;
		int lInputtingNeuron, lOutputtingNeuron;
		int lOutputtingNeuronIndex,lInputtingNeuronIndex;
		for (lOutputtingLayer=0; lOutputtingLayer<mLayerNeurons.length-1; lOutputtingLayer++){
			lInputtingLayer = lOutputtingLayer+1;
//			if (lLog) Log.getInstance().writeln("Layer "+lOutputtingLayer+"->"+lInputtingLayer);
			for (lInputtingNeuronIndex=0; lInputtingNeuronIndex<mLayerNeurons[lInputtingLayer].length; lInputtingNeuronIndex++){
				lInputtingNeuron = mLayerNeurons[lInputtingLayer][lInputtingNeuronIndex];
				mNeuronInput[lInputtingNeuron] = 0D;
				// BIAS bei allen Layern ausserhalb inputlayer bercksichtigen
				if (lOutputtingLayer>0){
					mNeuronInput[lInputtingNeuron] += mNeuronOutput[0] * mWeights[0][lInputtingNeuron];
//					if (lLog) Log.getInstance().writeln("BIAS->"+lInputtingNeuron+" ["+mNeuronOutput[0]+"]*w["+mWeights[0][lInputtingNeuron]+"]="+ (mNeuronOutput[0] * mWeights[0][lInputtingNeuron]) );
				}
				for (lOutputtingNeuronIndex=0; lOutputtingNeuronIndex<mLayerNeurons[lOutputtingLayer].length; lOutputtingNeuronIndex++){
					lOutputtingNeuron = mLayerNeurons[lOutputtingLayer][lOutputtingNeuronIndex];
					mNeuronInput[lInputtingNeuron] += mNeuronOutput[lOutputtingNeuron] * mWeights[lOutputtingNeuron][lInputtingNeuron];
//					if (lLog) Log.getInstance().writeln(lOutputtingNeuron+"->"+lInputtingNeuron+" ["+mNeuronOutput[lOutputtingNeuron]+"]*w["+mWeights[lOutputtingNeuron][lInputtingNeuron]+"]="+ (mNeuronOutput[lOutputtingNeuron] * mWeights[lOutputtingNeuron][lInputtingNeuron]) );
				}
//				if (lLog) Log.getInstance().writeln("in["+lInputtingNeuron+"]="+mNeuronInput[lInputtingNeuron]);
				mNeuronOutput[lInputtingNeuron] = mTransfer.eval(mNeuronInput[lInputtingNeuron]);
//				if (lLog) Log.getInstance().writeln("out["+lInputtingNeuron+"]="+mNeuronOutput[lInputtingNeuron]);
			}
		}
		
		// in output kopieren
		int [] outlayerneurons = mLayerNeurons[mLayerNeurons.length-1];
		for (int i=0; i<outlayerneurons.length; i++){
			lOutput[i] = mNeuronOutput[outlayerneurons[i]];
		}
		
		return lOutput;
	}
	
	/**
	 * Liefert die gesamte Anzahl der Neuronen in diesem Netz inklusive BIAS.
	 * 
	 * @return Die gesamte Anzahl der Neuronen in diesem Netz inklusive BIAS.
	 */
	public int getAmountNeurons(){
		return mAmountNeurons;
	}
	
	/**
	 * Liefert die Gre der Eingabeschicht (ohne BIAS).
	 * 
	 * @return Die Gre der Eingabeschicht (ohne BIAS).
	 */
	public int getInputLayerSize(){	
		return mLayerNeurons[0].length-1;
	}
	
	/**
	 * Liefert die Gre der Ausgabeschicht.
	 * 
	 * @return Die Gre der Ausgabeschicht.
	 */
	public int getOutputLayerSize(){
		return mLayerNeurons[mLayerNeurons.length-1].length;
	}
	
	/**
	 * Liefert die aktuelle Kantengewichtsmatrix.
	 * 
	 * @return Die aktuelle Kantengewichtsmatrix.
	 */
	public double [][] getWeights(){
		return mWeights;
	}
	
	/**
	 * Liefert die Indizes der Neuronen in einem Layer.<br/>
	 * Der erste Index des zurckgegebenen 2d-Arrays
	 * adressiert die einzelnen Schichten des Netzes.
	 * Das Netz hat also <code>getLayerNeurons().length</code>
	 * Schichten.<br/>
	 * Der zweite Index adressiert das entsprechende Neuron
	 * in der entsprechenden Schicht. Die gespeicherten Werte
	 * sind der Index des Neurons in der Gewichtematrix.<p/>
	 * Bsp. fr ein 2x1 Netz:<br/>
	 * <pre>
	 * getWeights()[getLayerNeurons[0][1]][getLayerNeurons[1][0]]
	 * </pre>
	 * liefert den Gewichtswert, der das erste Neuron der ersten Schicht (Eingabeschicht 
	 * hat den Index 0, erstes Neuron der Eingabeschicht mit Index 1)
	 * mit dem ersten Neuron der zweiten Schicht (hier Ausgabeschicht mit Index 1,
	 * erstes Neuron mit Index 0).
	 * Zu beachten ist, dass das Neuron mit dem Index 0 der Eingabeschicht das BIAS ist:<br/>
	 * <pre>
	 * getWeights()[getLayerNeurons[0][0]][getLayerNeurons[1][0]]
	 * </pre>
	 * liefert die Verbindung BIAS (Schicht 0, Neuronindex 0) zum
	 * ersten Neuron der zweiten Schicht (Schicht 1, Neuronindex 0).
	 * 
	 * @return Die Indizes der Neuronen in einem Layer.
	 */
	public int [][] getLayerNeurons(){
		return mLayerNeurons;
	}
	
	/**
	 * Liefert die Anzahl der verborgenen Schichten.
	 * 
	 * @return Die Anzahl der verborgenen Schichten.
	 */
	public int getHiddenLayerAmount(){
		return mLayerNeurons.length-2;
	}
	
	/* (non-Javadoc)
	 * @see observerPattern.Observable#getObserverManager()
	 */
	public ObserverManager getObserverManager() {
		return mObserverManager;
	}

	public String toString(){
		StringBuffer lSb = new StringBuffer();
		lSb.append(getClass().getName()+" "+getInputLayerSize()+"x");
		for (int i=0; i<getHiddenLayerAmount(); i++)
			lSb.append(mLayerNeurons[i+1].length+"x");
		lSb.append(getOutputLayerSize());
		return lSb.toString();
	}
	
	/**
	 * Liefert eine Zufallszahl im Bereich von [-pScaleFactor; +pScaleFactor].
	 *  
	 * @param pScaleFactor Ein positiver Wert, der festlegt, in welchem Bereich
	 * 						die Zufallszahl liegen soll.
	 * @return Zufallszahl im Bereich von [-pScaleFactor; +pScaleFactor].
	 */
	public static double getRandomWeightValue(double pScaleFactor){
		double lRand = getAbsRandomWeightValue(pScaleFactor);
		if (smRnd.nextBoolean()) lRand *= -1;
		return lRand;
	}
	
	/**
	 * Liefert eine positive Zufallszahl im Bereich von [0; +pScaleFactor].
	 * 
	 * @param pScaleFactor Eine positive Obergrenze fr die Zufallszahl.
	 * @return positive Zufallszahl im Bereich von [0; +pScaleFactor].
	 */
	public static double getAbsRandomWeightValue(double pScaleFactor){
		return smRnd.nextDouble() * pScaleFactor;
	}
	
	/**
	 * Berechnet den aktuellen Fehlergradienten fr die gegebenen Lerndatenstze.
	 * Es wird im Batch bzw. Offlineverfahren berechnet, bei der die Teil-
	 * Gradienten der einzelnen Lerndatenstze aufsummiert werden.
	 * 
	 * @param pNet Das Netz, dessen Fehlergradienten berechnet werden sollen.
	 * @param pInputPatterns Eingabemuster der Lerndatenstze.
	 * @param pOutputPatterns Ausgabemuster der Lerndatenstze.
	 * @return Aktuelle Fehlergradienten fr die gegebenen Lerndatenstze. 
	 * @throws PatternDoesNotMatchNetException Falls eines der gegebenen Lerndatenstze
	 * 		nicht der Eingabe-/ Ausgabeschichtgre des Netzes entspricht.
	 */
	public static double [][] calculateGradientsOffline(Net pNet, double [][] pInputPatterns, double [][] pOutputPatterns) throws PatternDoesNotMatchNetException{
		if (smLogMethodCall) Log.getInstance().writeln("Net.calculateGradients()");
//		boolean lLog;
/*		if (smResourceBundle.getString("showCalculateGradients").equalsIgnoreCase("true"))
			lLog = true;
		else lLog = false;
*/		
		double [] lNeuronOutput;
		double [][] lWeights = pNet.getWeights();
		double [][] lGradients = new double[lWeights.length][lWeights.length];
		double [] lNeuronError = new double[lWeights.length];
		int [][] lLayerNeurons = pNet.getLayerNeurons();
		int lLayerAmount = lLayerNeurons.length;
		int lLayerNeuron, lLayerNeuronIndex;
		int lOutputtingLayer, lInputtingLayer;
		int lOutputtingNeuronIndex, lInputtingNeuronIndex;
		int lOutputtingNeuron, lInputtingNeuron;
		if (pOutputPatterns[0].length != lLayerNeurons[lLayerAmount-1].length)
			throw new PatternDoesNotMatchNetException(
					smMSCResourceBundle.getString("OUTPUTPATTERN_LENGTH")+": "+pOutputPatterns[0].length+" "+
					smMSCResourceBundle.getString("OUTPUTLAYER_LENGTH")+": "+lLayerNeurons[lLayerAmount-1].length
				);
		
		// INIT GRADIENTS
		
		for (int i=0; i<pNet.getAmountNeurons(); i++)
			for (int j=0; j<pNet.getAmountNeurons(); j++)
				lGradients [i][j] = 0D;
		
		// CALCULATE NEURON ERRORS
		
		for (int lActualPattern=0; lActualPattern<pInputPatterns.length; lActualPattern++){
//			if (lLog){
//				Log.getInstance().writeln("");
//				Log.getInstance().writeln("Passing pattern "+lActualPattern);
//			}
			// calculate actual output
			lNeuronOutput = pNet.getNeuronsOutput(pInputPatterns[lActualPattern]);
			
			// calculate error for outputneurons
//			if (lLog) Log.getInstance().writeln("Error for outputlayer");
			for (lLayerNeuronIndex = 0; lLayerNeuronIndex<lLayerNeurons[lLayerAmount-1].length; lLayerNeuronIndex++){
				lLayerNeuron = lLayerNeurons[lLayerAmount-1][lLayerNeuronIndex];
//				if (lLog) Log.getInstance().write("Error neuron "+lLayerNeuron+": tanhdif(o["+lNeuronOutput[lLayerNeuron]+"])*"+
//					"(pat["+pOutputPatterns[lActualPattern][lLayerNeuronIndex]+"]-o["+lNeuronOutput[lLayerNeuron]+"])=");
				// tanhdif(o[x]) * (opat[x]-o[x])
				lNeuronError[lLayerNeuron] = 
					mTransfer.evalDif(lNeuronOutput[lLayerNeuron]) * 
					(pOutputPatterns[lActualPattern][lLayerNeuronIndex]-lNeuronOutput[lLayerNeuron]);
//				if (lLog) Log.getInstance().writeln(""+lNeuronError[lLayerNeuron]);
			}
			
			// calculate error for hiddenneurons
			// outputting im sinne des fehlers (bpg)
			for (lOutputtingLayer = lLayerAmount-1; lOutputtingLayer>=2; lOutputtingLayer--){
				lInputtingLayer = lOutputtingLayer-1;
//				if (lLog) Log.getInstance().writeln("Error feedback from layer "+lOutputtingLayer+" -> "+lInputtingLayer);
				for (lInputtingNeuronIndex = 0; lInputtingNeuronIndex<lLayerNeurons[lInputtingLayer].length; lInputtingNeuronIndex++){
					lInputtingNeuron = lLayerNeurons[lInputtingLayer][lInputtingNeuronIndex];
//					if (lLog) Log.getInstance().write("To neuron "+lInputtingNeuron+" <- from ");
					for (lOutputtingNeuronIndex = 0; lOutputtingNeuronIndex<lLayerNeurons[lOutputtingLayer].length; lOutputtingNeuronIndex++){
						lOutputtingNeuron = lLayerNeurons[lOutputtingLayer][lOutputtingNeuronIndex];
//						if (lLog) Log.getInstance().write(lOutputtingNeuron+": ");
						lNeuronError[lInputtingNeuron] += 
							lWeights[lInputtingNeuron][lOutputtingNeuron] *
							lNeuronError[lOutputtingNeuron];
					}
//					if (lLog) Log.getInstance().writeln("");
					lNeuronError[lInputtingNeuron] *= mTransfer.evalDif(lNeuronOutput[lInputtingNeuron]);
				}
			}
			
/*			if (lLog){ 
				Log.getInstance().writeln("Neuron errors"); 
				Log.getInstance().writeln(Formats.format(lNeuronError));
			}
*/			
			// CALCULATE WEIGHT GRADIENTS (OFFLINE->SUMMED)
			
			// gradw[1,4] = gradw[1,4] - e[4]*o[1]; // BIAS->hid
			// gradw[1,5] = gradw[1,5] - e[5]*o[1]; // BIAS->out
			// gradw[2,4] = gradw[2,4] - e[4]*o[2]; // in1->hid
			// gradw[3,4] = gradw[3,4] - e[4]*o[3]; // in2->hid
			// gradw[4,5] = gradw[4,5] - e[5]*o[4]; // hid->out
	
			for (lOutputtingLayer = 0; lOutputtingLayer<lLayerNeurons.length-1; lOutputtingLayer++){
				lInputtingLayer = lOutputtingLayer+1;
				
//				if (lLog) Log.getInstance().writeln("Calculating weightgradients layer "+lOutputtingLayer+"->"+lInputtingLayer);
				
				// BIAS bei allen Layern ausserhalb inputlayer bercksichtigen
				if (lOutputtingLayer>0){
					for (lInputtingNeuronIndex=0; lInputtingNeuronIndex<lLayerNeurons[lInputtingLayer].length; lInputtingNeuronIndex++){
						lInputtingNeuron = lLayerNeurons[lInputtingLayer][lInputtingNeuronIndex];
						lGradients[0][lInputtingNeuron] -=
							lNeuronError[lInputtingNeuron] * lNeuronOutput[0];
//						if (lLog) Log.getInstance().writeln("BIAS->"+lInputtingNeuron);
					}
				}
				
				for (lOutputtingNeuronIndex=0; lOutputtingNeuronIndex<lLayerNeurons[lOutputtingLayer].length; lOutputtingNeuronIndex++){
					lOutputtingNeuron = lLayerNeurons[lOutputtingLayer][lOutputtingNeuronIndex];
					for (lInputtingNeuronIndex=0; lInputtingNeuronIndex<lLayerNeurons[lInputtingLayer].length; lInputtingNeuronIndex++){
						lInputtingNeuron = lLayerNeurons[lInputtingLayer][lInputtingNeuronIndex];
						lGradients[lOutputtingNeuron][lInputtingNeuron] -=
							lNeuronError[lInputtingNeuron] * lNeuronOutput[lOutputtingNeuron];
//						if (lLog) Log.getInstance().writeln(lOutputtingNeuron+"->"+lInputtingNeuron);
					}
				}
			}

/*			if (lLog){ 
				Log.getInstance().writeln("Weight Gradients");
				Log.getInstance().writeln(Formats.format(lGradients));
			}
*/
		}
		return lGradients;
	}
	
	/**
	 * Wendet die Gewichtsaktualisierungen des Lernverfahrens an.
	 * Diese Methode ruft intern die {@link mscJNeuralNet.trainingAlgorithms.INetTrainingAlgorithm #applyWeightUpdate(int, int)}
	 * Methode des Lernverfahrens fr jedes zu aktualisierende Kantengewicht auf. 
	 * 
	 * @param pAlgo Das Lernverfahren, welches die Gewichtsaktualisierung berechnen soll.
	 */
	private static void updateWeights(INetTrainingAlgorithm pAlgo){
		if (smLogMethodCall) Log.getInstance().writeln("Net.updateWeights("+pAlgo+")");
/*		boolean lLog;
		if (smResourceBundle.getString("showUpdateWeights").equalsIgnoreCase("true"))
			lLog = true;
		else lLog = false;
*/		
		// CALCULATE & APPLY WEIGHT UPDATES
		
		// wb = wb - al*gradw - mu*gradwalt
		Net lNet = pAlgo.getNet();
		int lOutputtingLayer, lInputtingLayer;
		int lInputtingNeuronIndex, lOutputtingNeuronIndex;
		int lInputtingNeuron, lOutputtingNeuron;
		int [][] lLayerNeurons = lNet.getLayerNeurons();
		
		for (lOutputtingLayer = 0; lOutputtingLayer<lLayerNeurons.length-1; lOutputtingLayer++){
			lInputtingLayer = lOutputtingLayer+1;
			
//			if (lLog) Log.getInstance().writeln("Calculating weights for layer "+lOutputtingLayer+"->"+lInputtingLayer);
			
			// BIAS bei allen Layern ausserhalb inputlayer bercksichtigen
			if (lOutputtingLayer>0){
				for (lInputtingNeuronIndex=0; lInputtingNeuronIndex<lLayerNeurons[lInputtingLayer].length; lInputtingNeuronIndex++){
					lInputtingNeuron = lLayerNeurons[lInputtingLayer][lInputtingNeuronIndex];
					pAlgo.applyWeightUpdate(0, lInputtingNeuron);
//					if (lLog) Log.getInstance().writeln("BIAS->"+lInputtingNeuron);
				}
			}
			
			for (lOutputtingNeuronIndex=0; lOutputtingNeuronIndex<lLayerNeurons[lOutputtingLayer].length; lOutputtingNeuronIndex++){
				lOutputtingNeuron = lLayerNeurons[lOutputtingLayer][lOutputtingNeuronIndex];
				for (lInputtingNeuronIndex=0; lInputtingNeuronIndex<lLayerNeurons[lInputtingLayer].length; lInputtingNeuronIndex++){
					lInputtingNeuron = lLayerNeurons[lInputtingLayer][lInputtingNeuronIndex];
					pAlgo.applyWeightUpdate(lOutputtingNeuron, lInputtingNeuron);
//					if (lLog) Log.getInstance().writeln(lOutputtingNeuron+"->"+lInputtingNeuron);
				}
			}
		}
		
		lNet.getObserverManager().notifyObservers(lNet);
	}
	
	/**
	 * Fhrt einen Lernschritt mit dem Lernverfahren und dem im Lernverfahren vermerkten
	 * Netz aus, um die gegebenen Lerndatenstze zu trainieren.<br/>
	 * Vorher sollte das zu trainierende Netz durch den Aufruf der Methode 
	 * {@link mscJNeuralNet.trainingAlgorithms.INetTrainingAlgorithm#setNet(Net)}
	 * dem Lernverfahren mitgeteilt werden.<br/>
	 * Intern werden zuerst die Fehlergradienten des Netzes durch Aufruf der Methode 
	 * {@link #calculateGradientsOffline(Net, double[][], double[][])} berechnet.
	 * Dann werden die aktuellen Gradienten dem Lernverfahren durch Aufruf der Methode 
	 * {@link mscJNeuralNet.trainingAlgorithms.INetTrainingAlgorithm#setActualGradients(double[][])}
	 * mitgeteilt.
	 * Anschlieend wird die private Methode updateWeights(INetTrainingAlgorithm) aufgerufen,
	 * die schlielich ihrerseits fr jedes Kantengewicht die Methode 
	 * {@link mscJNeuralNet.trainingAlgorithms.INetTrainingAlgorithm#applyWeightUpdate(int, int)} 
	 * aufruft und den
	 * aktuellen Wert berechnet.<br/>
	 * Nach der Aktualisierung aller Gewichte wird durch die Methode 
	 * {@link mscJNeuralNet.trainingAlgorithms.INetTrainingAlgorithm#setCycle(int)}
	 * der aktuelle Lernschritt um eins erhht. 
	 * 
	 * @param pInputPatterns Eingabemuster der Lerndaten.
	 * @param pOutputPatterns Ausgabemuster der Lerndaten.
	 * @param pAlgo Lernverfahren, mit dem trainiert werden soll.
	 * @throws PatternDoesNotMatchNetException Falls eines der gegebenen Lerndatenstze
	 * 		nicht der Eingabe-/ Ausgabeschichtgre des Netzes entspricht.
	 */
	public static void train(double[][] pInputPatterns, double[][] pOutputPatterns, INetTrainingAlgorithm pAlgo)
		throws PatternDoesNotMatchNetException{
		
		if (smLogMethodCall) Log.getInstance().writeln("Net.train()");
		
		int [][] lLayerNeurons = pAlgo.getNet().getLayerNeurons();
		int lLayerAmount = lLayerNeurons.length;
		if (pOutputPatterns[0].length != lLayerNeurons[lLayerAmount-1].length)
			throw new PatternDoesNotMatchNetException(
					smMSCResourceBundle.getString("OUTPUTPATTERN_LENGTH")+": "+pOutputPatterns[0].length+" "+
					smMSCResourceBundle.getString("OUTPUTLAYER_LENGTH")+": "+lLayerNeurons[lLayerAmount-1].length
				);
		
		// CALCULATE NEURON ERRORS
		pAlgo.setActualGradients(Net.calculateGradientsOffline(pAlgo.getNet(), pInputPatterns, pOutputPatterns));

		// CALCULATE & APPLY WEIGHT UPDATES
		// wb = wb - al*gradw - mu*gradwalt
		updateWeights(pAlgo);
		
		pAlgo.setCycle(pAlgo.getCycle()+1);
	}
	
	/**
	 * Berechnet die Vorzeichenfunktion fr den gegebenen Wert.
	 * Die Vorzeichenfunktion ist definiert durch:<br/>
	 * <pre>
	 * 	sign(x) = -1   falls x < 0
	 * 	sign(x) = +1   falls x > 0
	 *  sign(x) =  0   falls x = 0
	 * </pre>
	 * 
	 * @param x Der Wert, zu dem die Vorzeichenfunktion berechnet werden soll.
	 * @return Die Vorzeichenfunktion fr den gegebenen Wert.
	 */
	public static double sign(double x){
		if (x>0) return 1D;
		else if (x<0) return -1D;
		return 0D;
	}
	
	/**
	 * Speichert das Netz in einer Datei.
	 * Das Netz wird als Text gespeichert, mit dem folgenden Format:
	 * <pre>
	 * Anzahl Layer
	 * Anzahl Neuronen Layer 1 (Eingabeschicht ohne BIAS)
	 * Anzahl Neuronen Layer 2
	 * ...
	 * Anzahl Neuronen Layer x (Ausgabeschicht)
	 * 
	 * Anzahl Neuronen insgesammt (inklusive BIAS)
	 * 
	 * Gewichtswert [0][0]
	 * Gewichtswert [0][1]
	 * ...
	 * Gewichtswert [0][x]
	 * Gewichtswert [1][0]
	 * Gewichtswert [1][1]
	 * ...
	 * Gewichtswert [1][x]
	 * ...
	 * Gewichtswert [x][x]
	 * </pre>
	 * @param pNet Das zu speichernde Netz.
	 * @param pTargetFile Datei, in der das Netz gespeichert werden soll.
	 * @throws IOException Falls bei der Ausgabe des Netzes in die Datei ein
	 * 						Fehler entsteht.
	 */
	public static void save(Net pNet, File pTargetFile) throws IOException{
		if (smLogMethodCall) Log.getInstance().writeln("Net.save("+pTargetFile+")");
		
		Log.getInstance().writeln("Saving "+pNet+" to file "+pTargetFile);
		
		PrintWriter lOut = new PrintWriter(new FileWriter(pTargetFile));
		save(pNet, lOut);
		lOut.close();
		
		Log.getInstance().writeln("Saving done.");
	}
	
	public static void save(Net pNet, PrintWriter pTargetOut) {
		double [][] lWeights = pNet.getWeights();
		int [][] lLayerNeurons = pNet.getLayerNeurons();
		
		// zuerst anzahl layer
		pTargetOut.print(Integer.toString(lLayerNeurons.length)+" ");
		// erster Layer ohne BIAS
		pTargetOut.print(Integer.toString(lLayerNeurons[0].length-1)+" ");
		// dann neuronen pro layer
		int x;
		for (x=1; x<lLayerNeurons.length; x++)
			pTargetOut.print(Integer.toString(lLayerNeurons[x].length)+" ");
		// dann neuronen insgesamt
		pTargetOut.println(Integer.toString(pNet.getAmountNeurons()));
		
		int y;
		for (x = 0; x<lWeights.length; x++){
			for (y = 0; y<lWeights[x].length; y++){
				pTargetOut.print(Double.toString(lWeights[x][y])+" ");
			}
			pTargetOut.println();
		}
		pTargetOut.flush();
	}
	
	public static Net load(Reader pIn) throws IOException {
		BufferedReader lIn = new BufferedReader(pIn);
		String lLine;
		StringBuffer lSbuffer = new StringBuffer();
		while ((lLine = lIn.readLine()) != null)
			lSbuffer.append(lLine+" ");
		lIn.close();
		
		String [] lTokens = lSbuffer.toString().split("\\s+");
		int lTokenIndex = 0;
		int [] lLayerSizes = new int[Integer.parseInt(lTokens[lTokenIndex++])];
		// Log.getInstance().writeln(" Found amount layers: "+Formats.format(lLayerSizes.length));
		int x,y;
		for (x=0; x<lLayerSizes.length; x++){
			lLayerSizes[x] = Integer.parseInt(lTokens[lTokenIndex++]);
			// Log.getInstance().writeln(" Found layersize "+x+" = "+Formats.format(lLayerSizes[x]));
		}
		
		// dann neuronen insgesamt
		int neuronsAtAll=Integer.parseInt(lTokens[lTokenIndex++]);
		// Log.getInstance().writeln(" Neurons at all (incl. BIAS): "+Double.parseDouble(lTokens[lTokenIndex++]));
		
		// Log.getInstance().writeln(" Reading data...");
		
		Net lNet = new Net(lLayerSizes);
		double [][] lWeights = lNet.getWeights();
		for (x=0; x<lWeights.length; x++){
			// Log.getInstance().write(" "+x+" ");
			for (y=0; y<lWeights[x].length; y++){
				lWeights[x][y] = Double.parseDouble(lTokens[lTokenIndex++]);
				// Log.getInstance().write(Formats.format(Formats.format(lWeights[x][y]),5 , 5)+" ");
			}
			// Log.getInstance().writeln("");
		}
		
		// Log.getInstance().writeln("Loaded net: "+lNet.toString());
		
//		if (smResourceBundle.getString("showLoadNet").equalsIgnoreCase("true"))
//			Log.getInstance().writeln(Formats.format(lWeights));
		
		return lNet;
	}
	
	/**
	 * Ldt ein Netz aus einer Datei.
	 * Das Format, welches geladen wird entspricht dem Speicherformat,
	 * welches unter {@link #save(Net, File)} beschrieben wurde.
	 * 
	 * @param pTargetFile Datei, aus der das Netz geladen werden soll. 
	 * @return Das geladene Netz.
	 * @throws IOException Falls bei der Ausgabe des Netzes in die Datei ein
	 * 						Fehler entsteht.
	 * @see #save(Net, File)
	 */
	public static Net load(File pTargetFile) throws IOException{
		if (smLogMethodCall) Log.getInstance().writeln("Net.load("+pTargetFile+")");
		Log.getInstance().writeln("Loading net from file "+pTargetFile);
		
		return load(new FileReader(pTargetFile));
	}
}
