package mscJNeuralNet.netPerformanceStatistics;

import java.util.ResourceBundle;

import formats.*;

import logger.Log;
import mscJNeuralNet.connectors.INetConnector;
import mscJNeuralNet.connectors.RandomSymmetryBreakingNetConnector;
import mscJNeuralNet.net.Net;
import mscJNeuralNet.net.PatternDoesNotMatchNetException;
import mscJNeuralNet.trainingAlgorithms.BackpropagationMomentum;
import mscJNeuralNet.trainingAlgorithms.INetTrainingAlgorithm;
import mscJNeuralNet.trainingAlgorithms.QuickProp;
import mscJNeuralNet.trainingAlgorithms.RProp;
import mscJNeuralNet.trainingAlgorithms.SuperSABMomentum;


/**
 * Diese Klasse dient der Erzeugung eines Berichtes ber die berechneten
 * Ist-Soll Fehlerwerte eines Netzes.
 * Auerdem steht eine Methode zur Verfgung, um ein Netz mehrfach von Grund auf
 * zu trainieren, um Statistiken ber die Performanz der gewhlten Parameter
 * des Netzes und des Lernverfahrens zu ermitteln.
 * <p/>
 * Beispiel zum Testen der verschiedenen Lernverfahren und ihrer Performanz auf dem
 * dreidimensionalen Xor-Problem. Es wird mit jedem Lernverfahren jeweils hundert mal
 * trainiert, wobei vor jedem Training die Kantengewichte erneut initialisiert werden.
 * Bei jedem Durchgang wird entweder bis zu einem avg. SSE Wert von 0.01 oder bis
 * 50 Lernschritte absolviert wurden trainiert.
 * <br/>
 * <pre>
 * // Lerndatenstze erzeugen 
 * double [][] trippleXORtrainInput = {
 *           {-1,-1,-1}, {-1,-1,1},
 *           {-1,1,-1}, {-1,1,1},
 *           {1,-1,-1}, {1,-1,1},
 *           {1,1,-1}, {1,1,1}};
 * double [][] trippleXORtrainOutput = {
 *           {-1},{1},
 *           {1},{-1},
 *           {1}, {-1},
 *           {-1}, {1}};
 * // Netzstruktur festlegen
 * int [] layerSizesTrippleXOr = {3, 3, 1};
 * 
 * // Parameter fr das Lernverfahren {@link mscJNeuralNet.trainingAlgorithms.BackpropagationMomentum}
 * // alpha, momentum
 * double [] momentum_1_bpg = {0.3, 0.4}; // best performance on tripplexor
 * 
 * // Parameter fr das Lernverfahren {@link mscJNeuralNet.trainingAlgorithms.SuperSABMomentum}
 * // momentum, alphaplus, alphaminus, alphainit
 * double [] momentum_1_ssab = {0.0};
 * 
 * // Parameter fr das Lernverfahren {@link mscJNeuralNet.trainingAlgorithms.QuickProp}
 * // alpha, max growth rate
 * double [] momentum_1_qprop = {0.3};
 * 
 * // Das zu traininerende Netz
 * Net lTrippleXorNet = new Net(layerSizesTrippleXOr);
 * 
 * // Die verschiedenen Lernverfahren
 * INetTrainingAlgorithm lAlgoBpgMom = new BackpropagationMomentum();
 * INetTrainingAlgorithm lAlgoSSAB = new SuperSABMomentum();
 * INetTrainingAlgorithm lAlgoRProp = new RProp();
 * INetTrainingAlgorithm lAlgoQProp = new QuickProp();
 * 
 * // Kantengewichtsinitialisierer
 * INetConnector lSymmNetConnector = new RandomSymmetryBreakingNetConnector();
 *
 * try{
 *   // Parameter fr Bpg festlegen
 *   lAlgoBpgMom.setLearningParameters(momentum_1_bpg);
 *   // Bpg hundert mal mit jeweils maximal 50 Lernschritten trainieren
 *   NetPerformanceReporter.getQualityReport(lTrippleXorNet, lAlgoBpgMom, trippleXORtrainInput, trippleXORtrainOutput, 50, 100, null, lSymmNetConnector);
 *   
 *   // Parameter fr SuperSAB festlegen
 *   lAlgoSSAB.setLearningParameters(momentum_1_ssab);
 *   // SuperSAB hundert mal mit jeweils maximal 50 Lernschritten trainieren
 *   NetPerformanceReporter.getQualityReport(lTrippleXorNet, lAlgoSSAB, trippleXORtrainInput, trippleXORtrainOutput, 50, 100, null, lSymmNetConnector);
 * 
 *   // RProp hundert mal mit jeweils maximal 50 Lernschritten trainieren
 *   NetPerformanceReporter.getQualityReport(lTrippleXorNet, lAlgoRProp, trippleXORtrainInput, trippleXORtrainOutput, 50, 100, null, lSymmNetConnector);
 *   
 *   // Parameter fr QProp festlegen
 *   lAlgoQProp.setLearningParameters(momentum_1_qprop);
 *   // QProp hundert mal mit jeweils maximal 50 Lernschritten trainieren
 *   NetPerformanceReporter.getQualityReport(lTrippleXorNet, lAlgoQProp, trippleXORtrainInput, trippleXORtrainOutput, 50, 100, null, lSymmNetConnector);
 * }
 * catch (Exception e){
 *   e.printStackTrace();
 * }
 * </pre> 
 * <p/>
 * Created on 31.05.2004
 * @author <A HREF="http://www.graviton.de">M. Serhat Cinar</A>
 * @version 13.06.2004
 * @see mscJNeuralNet.netPerformanceStatistics.NetPerformanceStatistics
 * @see mscJNeuralNet.netPerformanceStatistics.NetPerformanceStatisticsCalculator
 * @see mscJNeuralNet.netPerformanceStatistics
 */
public class NetPerformanceReporter {
	private static final ResourceBundle smResourceBundle = ResourceBundle.getBundle("debug");
	private static final boolean smLogMethodCall;
	private static final boolean smLogQualityReportDetails;
	private static final boolean smLogQualityReportSummary;
	static{
		if (smResourceBundle.getString("showMethodCall.NetStatisticsReporter").equalsIgnoreCase("true"))
			smLogMethodCall = true;
		else smLogMethodCall = false;
		
		if (smResourceBundle.getString("showQualityReportDetails.NetStatisticsReporter").equalsIgnoreCase("true"))
			smLogQualityReportDetails = true;
		else smLogQualityReportDetails = false;

		if (smResourceBundle.getString("showQualityReportSummary.NetStatisticsReporter").equalsIgnoreCase("true"))
			smLogQualityReportSummary = true;
		else smLogQualityReportSummary = false;
	}
	
	/**
	 * Ein Netz wird mehrfach von Grund auf trainiert, um Statistiken ber 
	 * die Performanz der gewhlten Parameter des Netzes und des Lernverfahrens zu ermitteln.
	 * Es wird mit dem gegebenen Lernverfahren jeweils die angegebene Anzahle mal
	 * trainiert, wobei vor jedem Training die Kantengewichte mit dem Konnektor erneut 
	 * initialisiert werden.
	 * Bei jedem Durchgang wird entweder bis zu einem avg. SSE Wert von 0.01 oder bis
	 * die angegebene Anzahl der Lernschritte absolviert wurden trainiert.
	 * Nach jedem Aufruf wird ein Bericht ausgegeben.
	 * 
	 * @param pNet Das Netz, welches zu testen ist.
	 * @param pAlgo Das zu benutzende Lernverfahren. 
	 * @param pInputPattern Die Eingabemuster fr das Training.
	 * @param pOutputPattern Die Ausgabemuster fr das Training (Soll-Werte).
	 * @param pRoundsToTest Anzahl der Testdurchlufe.
	 * @param pMaxCycles Anzahl der maximalen Lernschritte pro Testdurchlauf.
	 * @param pConnectorParameters Parameter fr den INetConnector.
	 * @param pNetConnectionAlgo Der Kantengewichtsinitialisierungsalgorithmus.
	 * @throws PatternDoesNotMatchNetException Falls eines der gegebenen Lerndatenstze
	 * 		nicht der Eingabe-/ Ausgabeschichtgre des Netzes entspricht.
	 */
	public static void getQualityReport(Net pNet, INetTrainingAlgorithm pAlgo, 
		double[][] pInputPattern, double[][]pOutputPattern, 
		int pRoundsToTest, int pMaxCycles, 
		double [] pConnectorParameters, INetConnector pNetConnectionAlgo)
		throws PatternDoesNotMatchNetException{
			
		if (smLogMethodCall) Log.getInstance().writeln("NetStatisticsReporter.getQualityReport().");
		
		NetPerformanceStatistics lNetStat = NetPerformanceStatisticsCalculator.calculateErrors(pNet, pInputPattern, pOutputPattern);
		// NetStatisticsCalculator.NetStatistics lNetStat;
		int lGood=0;
		int lBad=0;
		double avg = 0;
		double avgcy = 0;
		int avgcyamount = 0;
		StringBuffer lSbAllResults = new StringBuffer();
		if (smLogQualityReportDetails || smLogQualityReportSummary){ 
			Log.getInstance().writeln("Testing "+pAlgo);
			Log.getInstance().writeln("Connector Parameters: "+Formats.format(pConnectorParameters));
			Log.getInstance().writeln("WeightConnector: "+pNetConnectionAlgo);
			Log.getInstance().write("Net layers:");
			for (int i=0; i<pNet.getLayerNeurons().length-1; i++)
				Log.getInstance().write(pNet.getLayerNeurons()[i].length+"x");
			Log.getInstance().writeln(pNet.getLayerNeurons()[pNet.getLayerNeurons().length-1].length+"");
		}
		if (smLogQualityReportDetails){
			Log.getInstance().writeln("Input:");
			Log.getInstance().writeln(Formats.format(pInputPattern));
			Log.getInstance().writeln("Output:");
			Log.getInstance().writeln(Formats.format(pOutputPattern));
		}
		for (int lTry=0; lTry<pRoundsToTest; lTry++){
			pAlgo.setNet(pNet);
			pNetConnectionAlgo.connectNet(pNet, pConnectorParameters);
			if (smLogQualityReportDetails){
				Log.getInstance().writeln("Test Nr.: "+lTry+" ");
				Log.getInstance().writeln("Net:");
				Log.getInstance().writeln(Formats.format(pNet.getWeights()));
			}
			int lCycle = 0;
			while (++lCycle<pMaxCycles){
				Net.train(pInputPattern, pOutputPattern, pAlgo);
				lNetStat = NetPerformanceStatisticsCalculator.calculateErrors(pNet, pInputPattern, pOutputPattern);
				if (lNetStat.averageSumOfSquaredErrors<=0.01) break;
			}
			if (lNetStat.averageSumOfSquaredErrors<=0.01){
				lGood++;
				if (smLogQualityReportDetails){
					Log.getInstance().write("(SSE:"+Formats.format(lNetStat.averageSumOfSquaredErrors)+"/"+" ABSE:"+Formats.format(lNetStat.averageSumOfAbsoluteErrors)+")+@"+lCycle+"  ");
				}
				lSbAllResults.append(lTry+": (SSE:"+Formats.format(lNetStat.averageSumOfSquaredErrors)+"/"+" ABSE:"+Formats.format(lNetStat.averageSumOfAbsoluteErrors)+")+@"+lCycle+"\n");
				avgcyamount++;
				avgcy += lCycle;
			}
			else{
				lBad++;
				if (smLogQualityReportDetails){
					Log.getInstance().write("(SSE:"+Formats.format(lNetStat.averageSumOfSquaredErrors)+"/"+" ABSE:"+Formats.format(lNetStat.averageSumOfAbsoluteErrors)+") @"+lCycle+"  ");
				}
				lSbAllResults.append(lTry+": (SSE:"+Formats.format(lNetStat.averageSumOfSquaredErrors)+"/"+" ABSE:"+Formats.format(lNetStat.averageSumOfAbsoluteErrors)+") @"+lCycle+"\n");
			}
			if (smLogQualityReportDetails)
				Log.getInstance().writeln(getNetPerformance(lNetStat, pAlgo.getCycle()));
			
			avg += lNetStat.averageSumOfSquaredErrors;
		}
		avg /= pRoundsToTest;
		avgcy /= avgcyamount;
		if (smLogQualityReportDetails || smLogQualityReportSummary){
			Log.getInstance().writeln("All results: ");
			Log.getInstance().write(lSbAllResults.toString());
			Log.getInstance().writeln(pRoundsToTest+" Tests with"+
				" Good: "+(((double)lGood)/((double)pRoundsToTest))+
				" Bad: "+(((double)lBad)/((double)pRoundsToTest))+
				" AverageSSE: "+Formats.format(avg)+
				" AverageSuccessCycle: "+Formats.format(avgcy));
			Log.getInstance().writeln("");
		}
	}
	
	/**
	 * Generiert einen Bericht ber die Performanz eines Netzes. 
	 * Der Bericht zeigt die durchschnittlichen Fehlerwerte auf den 
	 * Lerndatenstzen sowie den absoluten Fehler fr jedes Ausgabeneuron (Ist-Wert)
	 * zu den Ausgabemustern (Soll-Wert).<br/>
	 * Die erste Zeile des Berichtes enthlt Informationen ber die durchschnittliche
	 * Performaz des Netzes auf allen Lernverfahren:<br/>
	 * <pre>
	 * avg.ABSE:0.03 / avg.SSE:0 / avg.RMSE:0.01 / max. Neuron AbsE:0.03 @[Lernschritt]
	 * </pre>
	 * <p/>
	 * Fr jeden Lerndatensatz wird im Bericht eine eigene Zeile produziert.<br/>
	 * Das Muster fr jede Zeile hat die Form:<br/>
	 * <pre>
	 * Pattern [pattern number]  
	 * [output pattern value]=[netoutput] AbsE:[absolute neuron error]
	 * </pre>
	 * <p/>
	 * Der Auszug unten ist ein Bericht ber ein Netz mit drei Ausgabeneuronen,
	 * welches mit 16 Lerndatenstzen trainiert wurde.<br/>
	 * <pre>
	 * avg.ABSE:0.03 / avg.SSE:0 / avg.RMSE:0.01 / max. Neuron AbsE:0.03 @377
	 * Pattern 0
	 * 0.6=0.61 (AbsE:0.01)  0.1=0.08 (AbsE:0.02)  0.2=0.2 (AbsE:0)  
	 * Pattern 1
	 * 0.7=0.69 (AbsE:0.01)  0.1=0.11 (AbsE:0.01)  0.2=0.2 (AbsE:0)  
	 * Pattern 2
	 * 0.1=0.11 (AbsE:0.01)  0.2=0.18 (AbsE:0.02)  0.2=0.19 (AbsE:0.01)  
	 * Pattern 3
	 * 0.1=0.1 (AbsE:0)  0.2=0.19 (AbsE:0.01)  0.2=0.19 (AbsE:0.01)  
	 * Pattern 4
	 * 0.7=0.67 (AbsE:0.03)  0.1=0.09 (AbsE:0.01)  0.2=0.2 (AbsE:0)  
	 * Pattern 5
	 * 0.6=0.61 (AbsE:0.01)  0.1=0.1 (AbsE:0)  0.2=0.2 (AbsE:0)  
	 * Pattern 6
	 * 0.5=0.49 (AbsE:0.01)  0.1=0.09 (AbsE:0.01)  0.2=0.19 (AbsE:0.01)  
	 * Pattern 7
	 * 0.4=0.4 (AbsE:0)  0.1=0.11 (AbsE:0.01)  0.2=0.22 (AbsE:0.02)  
	 * Pattern 8
	 * 0.3=0.27 (AbsE:0.03)  0.1=0.12 (AbsE:0.02)  0.2=0.21 (AbsE:0.01)  
	 * Pattern 9
	 * 0.2=0.21 (AbsE:0.01)  0.1=0.09 (AbsE:0.01)  0.2=0.21 (AbsE:0.01)  
	 * Pattern 10
	 * 0.1=0.12 (AbsE:0.02)  0.1=0.12 (AbsE:0.02)  0.2=0.2 (AbsE:0)  
	 * Pattern 11
	 * 0.1=0.1 (AbsE:0)  0.1=0.1 (AbsE:0)  0.2=0.2 (AbsE:0)  
	 * Pattern 12
	 * 0.2=0.2 (AbsE:0)  0.1=0.09 (AbsE:0.01)  0.2=0.2 (AbsE:0)  
	 * Pattern 13
	 * 0.3=0.3 (AbsE:0)  0.1=0.11 (AbsE:0.01)  0.2=0.21 (AbsE:0.01)  
	 * Pattern 14
	 * 0.4=0.4 (AbsE:0)  0.1=0.1 (AbsE:0)  0.2=0.19 (AbsE:0.01)  
	 * Pattern 15
	 * 0.5=0.51 (AbsE:0.01)  0.1=0.1 (AbsE:0)  0.2=0.2 (AbsE:0)
	 * </pre>
	 * 
	 * @param pNetStat Das NetPerformanceStatistics Objekt, zu dem der Bericht erstellt werden soll.
	 * @param pLearningCycle Der Lernschritt, zu dem das NetPerformanceStatistics Objekt generiert wurde.
	 * @return String, der den Bericht als formatierten Text enthlt.
	 */
	public static String getNetPerformance(NetPerformanceStatistics pNetStat, int pLearningCycle){
		StringBuffer lSb = new StringBuffer();
		lSb.append(
			"avg.ABSE:"+pNetStat.averageSumOfAbsoluteErrors+" /"+
			" avg.SSE:"+pNetStat.averageSumOfSquaredErrors+" /"+
			" avg.RMSE:"+pNetStat.averageRootMeanSquareError+" /"+
			" max. Neuron AbsE:"+pNetStat.maximumNeuronAbsoluteError+
			" @"+pLearningCycle+"\n");
		for (int lActualPattern=0; lActualPattern<pNetStat.inputPatterns.length; lActualPattern++){
			lSb.append("Pattern "+lActualPattern+"\n");
			for (int lActualNeuron=0; lActualNeuron<pNetStat.outputPatterns[lActualPattern].length; lActualNeuron++){
				lSb.append(Formats.format(pNetStat.outputPatterns[lActualPattern][lActualNeuron])+"=");
				lSb.append(Formats.format(pNetStat.netOutputs[lActualPattern][lActualNeuron])+" ");
				lSb.append("(AbsE:"+Formats.format(pNetStat.neuronAbsoluteError[lActualPattern][lActualNeuron])+")  ");
			}
			lSb.append("\n");
		}
		return lSb.toString();
	}
}
