package mscJNeuralNet.netPerformanceStatistics;
import java.util.ResourceBundle;

import logger.Log;
import mscJNeuralNet.net.Net;
import mscJNeuralNet.net.PatternDoesNotMatchNetException;

/**
 * Diese Klasse berechnet die aktuellen Ist-Soll Fehlerwerte
 * eines Netzes.
 * <p/>
 * Created on 04.04.2004
 * @author <A HREF="http://www.graviton.de">M. Serhat Cinar</A>
 * @version 05.06.2004
 * @see mscJNeuralNet.netPerformanceStatistics
 * @see mscJNeuralNet.netPerformanceStatistics.NetPerformanceStatistics
 */
public class NetPerformanceStatisticsCalculator {
	
	private static final ResourceBundle smResourceBundle = ResourceBundle.getBundle("debug");
	private static final boolean smLogMethodCall;
	private static final boolean smLogCalcErr;
	static{
		if (smResourceBundle.getString("showMethodCall.NetStatisticsCalculator").equalsIgnoreCase("true"))
			smLogMethodCall = true;
		else smLogMethodCall = false;
		if (smResourceBundle.getString("showCalculateError.NetStatisticsCalculator").equalsIgnoreCase("true"))
			smLogCalcErr = true;
		else smLogCalcErr = false;
	}
	
	private NetPerformanceStatisticsCalculator(){/** tool */}
	
	// TODO pattern-version 
	/**
	 * Berechnet alle aktuellen Ist-Soll Fehlerwerte eines Netzes und gibt diese
	 * als {@link NetPerformanceStatistics} Objekt zurck.
	 * 
	 * @param pNet Das Netz, mit dem die Berechnungen erstellt werden sollen.
	 * @param pInputPatterns Eingabemuster, welche dem Netz prsentiert werden.
	 * @param pOutputPatterns Ausgabemuster, die das Netz lernen soll (Soll-Werte).
	 * @return Alle aktuell berechneten Ist-Soll Fehlerwerte des Netzes.
	 * @throws PatternDoesNotMatchNetException Falls eines der gegebenen Lerndatenstze
	 * 		nicht der Eingabe-/ Ausgabeschichtgre des Netzes entspricht.
	 */
	public static NetPerformanceStatistics calculateErrors(Net pNet, double[][] pInputPatterns, double[][] pOutputPatterns) throws PatternDoesNotMatchNetException{
		if (smLogMethodCall) Log.getInstance().writeln("Netstat.calculateErrors()");
		NetPerformanceStatistics lNetErrors = new NetPerformanceStatistics();
		lNetErrors.inputPatterns = pInputPatterns;
		lNetErrors.outputPatterns = pOutputPatterns;
		
		// netzausgaben auf verschiedene patterns berechnen 
		lNetErrors.netOutputs = new double[pInputPatterns.length][];
		for (int lPatternIndex = 0; lPatternIndex<pInputPatterns.length; lPatternIndex++){
			lNetErrors.netOutputs[lPatternIndex] = pNet.getNetOutput(pInputPatterns[lPatternIndex]);
		}
		
		// fehlerwerte initialisieren
		lNetErrors.sumOfAbsoluteErrors = new double[pInputPatterns.length];
		lNetErrors.sumOfSquaredErrors = new double[pInputPatterns.length];
		lNetErrors.rootMeanSquareErrors = new double[pInputPatterns.length];
		lNetErrors.neuronAbsoluteError = new double[pInputPatterns.length][lNetErrors.netOutputs[0].length];
		lNetErrors.neuronSquaredError = new double[pInputPatterns.length][lNetErrors.netOutputs[0].length];
		lNetErrors.averageSumOfAbsoluteErrors = 0D;
		lNetErrors.averageSumOfSquaredErrors = 0D;
		lNetErrors.averageRootMeanSquareError = 0D;
		lNetErrors.maximumNeuronAbsoluteError = -1D;
		lNetErrors.maximumNeuronAbsoluteErrorNeuronIndex = 0;
		lNetErrors.maximumNeuronAbsoluteErrorPatternIndex = 0;
		
		int lActualOutputNeuron;
		double lDifference;
		
		// schleife alle patterns
		for (int lActualPattern = 0; lActualPattern<pInputPatterns.length; lActualPattern++){
			// summe der absolutbetrge der fehler ber aktuelles pattern
			lNetErrors.sumOfAbsoluteErrors[lActualPattern] = 0;
			lNetErrors.sumOfSquaredErrors[lActualPattern] = 0;
			
			if (smLogCalcErr)
				Log.getInstance().writeln("Pat "+lActualPattern+" ");
			
			// schleife alle ausgabeneuronen
			for (lActualOutputNeuron=0; lActualOutputNeuron<lNetErrors.netOutputs[lActualPattern].length; lActualOutputNeuron++){
				// abstand beider werte
				lDifference = pOutputPatterns[lActualPattern][lActualOutputNeuron] - lNetErrors.netOutputs[lActualPattern][lActualOutputNeuron];
				// fehler fr neuron
				lNetErrors.neuronAbsoluteError[lActualPattern][lActualOutputNeuron] = Math.abs(lDifference);
				// grsster fehler fr ein einzelnes neuron
				if (lNetErrors.neuronAbsoluteError[lActualPattern][lActualOutputNeuron] > lNetErrors.maximumNeuronAbsoluteError){
					lNetErrors.maximumNeuronAbsoluteError = lNetErrors.neuronAbsoluteError[lActualPattern][lActualOutputNeuron];
					lNetErrors.maximumNeuronAbsoluteErrorNeuronIndex = lActualOutputNeuron;
					lNetErrors.maximumNeuronAbsoluteErrorPatternIndex = lActualPattern;
				}
				// summe der fehler fr ein pattern
				lNetErrors.sumOfAbsoluteErrors[lActualPattern] += lNetErrors.neuronAbsoluteError[lActualPattern][lActualOutputNeuron];
				// fehler fr neuron
				lNetErrors.neuronSquaredError[lActualPattern][lActualOutputNeuron] = Math.pow(lDifference, 2.0);
				// summe der fehler fr ein pattern
				lNetErrors.sumOfSquaredErrors[lActualPattern] += lNetErrors.neuronSquaredError[lActualPattern][lActualOutputNeuron];
				
				if (smLogCalcErr){ 
					Log.getInstance().write("Neuron "+lActualOutputNeuron+": ");
					Log.getInstance().writeln(
						"("+pOutputPatterns[lActualPattern][lActualOutputNeuron]+")-("+
						lNetErrors.netOutputs[lActualPattern][lActualOutputNeuron]+")="+
						lDifference+
						" (ABSE: "+lNetErrors.neuronAbsoluteError[lActualPattern][lActualOutputNeuron]+")"+
						" (SSE: "+lNetErrors.sumOfSquaredErrors[lActualPattern]+")");
				}
			} // ende schleife alle ausgabeneuronen
			
			lNetErrors.rootMeanSquareErrors[lActualPattern] = 
				Math.sqrt(
					lNetErrors.sumOfSquaredErrors[lActualPattern]/
					lNetErrors.netOutputs[lActualPattern].length);
			
			lNetErrors.averageSumOfAbsoluteErrors += lNetErrors.sumOfAbsoluteErrors[lActualPattern];
			lNetErrors.averageSumOfSquaredErrors += lNetErrors.sumOfSquaredErrors[lActualPattern];
			lNetErrors.averageRootMeanSquareError += lNetErrors.rootMeanSquareErrors[lActualPattern];
		}// ende schleife alle patterns
		
		lNetErrors.averageSumOfAbsoluteErrors /= pInputPatterns.length;
		lNetErrors.averageSumOfSquaredErrors /= pInputPatterns.length;
		lNetErrors.averageRootMeanSquareError /= pInputPatterns.length;
 		
		return lNetErrors;
	}
}
