package mscJNeuralNet.trainingAlgorithms;

import java.util.ResourceBundle;

import mscJNeuralNet.net.Net;

/**
 * Das Quick Propagation Lernverfahren nach Fahlman.
 * Implentiert nach der Arbeit
 * <i>Scott E. Fahlman, An Empirical Study of Learning Speed in Back-Propagation Networks, 1988</i>
 * und der Internetseite 
 * <a href="http://www.cs.ncl.ac.uk/modules/1998-99/csc325/projects/quickprop/implemen.htm">
 * Description and Implementation of the Algorithm</a> 
 * <p/>
 * Das Quickprop Lernverfahren benutzt das Newtonsche Abstiegsverfahren zur Ermittlung
 * eines Minimums und ist daher recht schnell.
 * <p/>
 * Die Aktualisierung der Kantengewichte wird wie folgt berechnet:<br/>
 * (Auszug aus <i>Prof. Dr. H. Westenberger, Skript Neuronale Netze 99V011</i>)
 * <pre>
 * 		Gradiententerm
 * 	    G = - alpha * Gradient Wij falls Gradient Wij = 0 oder 
 *                                       Delta WijAlt = 0 oder 
 *                                       sign(Gradient Wij) == sign(Gradient WijAlt)
 *      G = 0 sonst
 * 
 *      Parabelterm
 *      P = 0 falls getCycle() == 0 oder 
 *                  Delta WijAlt == 0 oder 
 *                  Math.abs(Gradient Wij) > Math.abs(Gradient WijAlt)
 *            sonst
 *      P = Math.min(MaxGrowthRate, (Gradient Wij / (Gradient WijAlt-Gradient Wij))) * Delta WijAlt
 *
 *      Delta Wij = G + P
 * </pre>
 * Der normalerweise bliche Momentumterm fr G (hier normales Backpropagation)
 * wurde nicht implementiert.
 * <p/>
 * Created on 03.06.2004
 * @author <A HREF="http://www.graviton.de">M. Serhat Cinar</A>
 * @version 05.06.2004
 * @see mscJNeuralNet.trainingAlgorithms.INetTrainingAlgorithm
 * @see mscJNeuralNet.trainingAlgorithms.AbstractTrainingAlgorithm
 * @see mscJNeuralNet.trainingAlgorithms
 */
public class QuickProp extends AbstractTrainingAlgorithm {
	private static final ResourceBundle smMSCResourceBundle = ResourceBundle.getBundle("mscJNeuronalNet_Text");
	
	private double [][] mWeightChangeOld;
	private double [][] mWeightChange;
	private double [][] mGradientWOld;

	private static final String smParameterNames [] = {
		smMSCResourceBundle.getString("ALPHA_LEARNING_RATE"),
		smMSCResourceBundle.getString("MAX_GROWTH_RATE")
	};
	private static final double [] smParameterDefaultValues = {1.75, 1D};

	// mu
	private double mMaxGrowth = smParameterDefaultValues[0];
	private double mAlpha = smParameterDefaultValues[1];
		
	/* (non-Javadoc)
	 * @see mscJNeuralNet.trainingAlgorithms.INetTrainingAlgorithm#init()
	 */
	public void init(){
		super.init();

		// init gradients
		mGradientWOld = new double [getNet().getAmountNeurons()][getNet().getAmountNeurons()];
		mWeightChange = new double [getNet().getAmountNeurons()][getNet().getAmountNeurons()];
		mWeightChangeOld = new double [getNet().getAmountNeurons()][getNet().getAmountNeurons()];
		for (int x=0; x<getNet().getAmountNeurons(); x++){
			for (int y=0; y<getNet().getAmountNeurons(); y++){
				mGradientWOld[x][y] = 0D;
				mWeightChange[x][y] = 0D;
				mWeightChangeOld[x][y] = 0D;
			}
		}
		setCycle(0);
	}
	
	private double mP;
	private double mG;
	private double mMu;
	private double mActualGradient;
	private double mOldGradient;
	private double mActualWeightChange;

	/* (non-Javadoc)
	 * @see mscJNeuralNet.trainingAlgorithms.INetTrainingAlgorithm#calculateWeightUpdate(int, int)
	 */
	public void applyWeightUpdate(int pOutputting, int pInputting){
		super.applyWeightUpdate(pOutputting, pInputting);
		
		// g(t)
		mActualGradient = getActualGradients()[pOutputting][pInputting];
		mOldGradient = mGradientWOld[pOutputting][pInputting];
		mActualWeightChange = mWeightChange[pOutputting][pInputting];
		
		mG = 0D;
		if (mActualGradient == 0 || mActualWeightChange == 0 || Net.sign(mActualGradient) == Net.sign(mOldGradient)){
			mG = -1D*mAlpha*mActualGradient;
		}
		
		// p(t)
		if (getCycle() == 0 || mActualWeightChange == 0 || Math.abs(mActualGradient) > Math.abs(mOldGradient)){
			mP = 0D;
		}
		else{
			mMu = mActualGradient / (mOldGradient-mActualGradient);
			mP =  Math.min(mMaxGrowth, mMu) * mActualWeightChange;
		}
		
		// weightchange = g(t) + p(t)
		mWeightChange[pOutputting][pInputting] = mG+mP;
		
		getNet().getWeights()[pOutputting][pInputting] += mWeightChange[pOutputting][pInputting];
		mWeightChangeOld[pOutputting][pInputting] = mWeightChange[pOutputting][pInputting];
		mGradientWOld[pOutputting][pInputting] = mActualGradient;
	}

	/**
	 * Dieses Lernverfahren untersttzt folgende Parameter: <br/>
	 * <pre>
	 * Index  Bezeichnung   Funktion
	 *   0    Alpha         Lernschrittweite
	 *   1    MaxGrowthRate Beschrnkt die gemachten Schrittweiten als Maximum.
	 * </pre>
	 * Am Ende der Methode werden die Observer benachrichtigt.
	 *  
	 * @see INetTrainingAlgorithm#setLearningParameters(double[])
	 */
	public void setLearningParameters(double[] pLearningParameters) {
		if (pLearningParameters.length>0){
			if (pLearningParameters[0] >= 0D) mAlpha = pLearningParameters[0];
			if (pLearningParameters.length>1){
				if (pLearningParameters[1] >= 0D) mMaxGrowth = pLearningParameters[1];
			}
		}
		getObserverManager().notifyObservers(this);
	}
	
	/**
	 * Dieses Lernverfahren untersttzt folgende Parameter: <br/>
	 * <pre>
	 * Index  Bezeichnung   Funktion
	 *   0    Alpha         Lernschrittweite
	 *   1    MaxGrowthRate Beschrnkt die gemachten Schrittweiten als Maximum.
	 * </pre>
	 *  
	 * @see mscJNeuralNet.trainingAlgorithms.INetTrainingAlgorithm#getLearningParameters()
	 */
	public double[] getLearningParameters() {
		double [] lParameters = new double[smParameterNames.length];
		lParameters[0] = mAlpha;
		lParameters[1] = mMaxGrowth;
		return lParameters;
	}

	/**
	 * Dieses Lernverfahren untersttzt folgende Parameter: <br/>
	 * <pre>
	 * Index  Bezeichnung   Funktion
	 *   0    Alpha         Lernschrittweite
	 *   1    MaxGrowthRate Beschrnkt die gemachten Schrittweiten als Maximum.
	 * </pre>

	 * @see mscJNeuralNet.trainingAlgorithms.INetTrainingAlgorithm#getDefaultLearningParameters()
	 */
	public double [] getDefaultLearningParameters(){
		return smParameterDefaultValues;
	}
	
	/**
	 * Dieses Lernverfahren untersttzt folgende Parameter: <br/>
	 * <pre>
	 * Index  Bezeichnung   Funktion
	 *   0    Alpha         Lernschrittweite
	 *   1    MaxGrowthRate Beschrnkt die gemachten Schrittweiten als Maximum.
	 * </pre>
	 * 
	 * @see mscJNeuralNet.trainingAlgorithms.INetTrainingAlgorithm#getLearningParameterNames()
	 */
	public String[] getLearningParameterNames() {
		return smParameterNames;
	}
	
	/**
	 * Liefert den String "QuickProp" zurck.
	 * 
	 * @see mscJNeuralNet.trainingAlgorithms.INetTrainingAlgorithm#getAlgorithmName()
	 */
	public String getAlgorithmName(){
		return "QuickProp";
	}
	
	public String toString(){
		return "QuickProp alpha: "+mAlpha+" MaxGrowthRate: "+mMaxGrowth;
	}
}
