package mscJNeuralNet.trainingAlgorithms;

import java.util.ResourceBundle;

import formats.Formats;

import logger.Log;

/**
 * 
 * Das SuperSAB Lernverfahren nach Tollenaere mit Momentum-Term.<br/>
 * Implentiert nach der Arbeit 
 * <i>T. Tollenaere - Supersab: Fast adaptive backpropagation with good scaling properties, 1990</i>
 * <p/>
 * Das SuperSAB Verfahren erweitert das Backpropagation Verfahren um
 * variable Lernschrittweiten (Alpha ij) fr jedes einzelne Gewicht.
 * Diese Version wurde um einen Momentum-Term erweitert.
 * <p/>
 * Die Aktualisierung der Kantengewichte wird wie folgt berechnet:<br/>
 * <pre>
 *   Gradientenrichtung = Gradient Wij * Gradient WijAlt
 *   
 *   Falls Gradientenrichtung > 0
 *     Alpha ij *= Alpha Plus
 *   Falls Gradientenrichtung < 0
 *     Alpha ij *= Alpha Minus
 * 
 *                SuperSAB-Term                   Momentum-Term
 *   Delta Wij = -1 * Alpha ij * Gradient Wij + Momentum * Delta WijAlt
 * </pre>
 * <p/>
 * Created on 02.06.2004
 * @author <A HREF="http://www.graviton.de">M. Serhat Cinar</A>
 * @version 05.06.2004
 * @see mscJNeuralNet.trainingAlgorithms.INetTrainingAlgorithm
 * @see mscJNeuralNet.trainingAlgorithms.AbstractTrainingAlgorithm
 * @see mscJNeuralNet.trainingAlgorithms
 */
public class SuperSABMomentum extends AbstractTrainingAlgorithm{
	private static final ResourceBundle smMSCResourceBundle = ResourceBundle.getBundle("mscJNeuronalNet_Text");
	private static final ResourceBundle smResourceBundle = ResourceBundle.getBundle("debug");
//	private static final boolean smLogCalculateWeightUpdate;
//	private static final boolean smLogTrain;
	private static final boolean smLogMethodCall;
	static{
//		if (smResourceBundle.getString("showLearningCalculateWeightUpdateSSAB").equalsIgnoreCase("true"))
//			smLogCalculateWeightUpdate = true;
//		else smLogCalculateWeightUpdate = false;
		if (smResourceBundle.getString("showMethodCall.TrainingAlgorithm").equalsIgnoreCase("true"))
			smLogMethodCall = true;
		else smLogMethodCall = false;
//		if (smResourceBundle.getString("showTrainSSAB").equalsIgnoreCase("true"))
//			smLogTrain = true;
//		else smLogTrain = false;
	}

	private double [][] mGradientWOld;
	private double [][] mAlphaIJ;
	private double [][] mWeightChange;
	private double [][] mWeightChangeOld;
	
	private static final String smParameterNames [] = {
		smMSCResourceBundle.getString("MOMENTUM"),
		smMSCResourceBundle.getString("ALPHA_PLUS"),
		smMSCResourceBundle.getString("ALPHA_MINUS"), 
		smMSCResourceBundle.getString("ALPHA_ZERO")
	};
	private static final double [] smParameterDefaultValues = {0D, 1.1, .5, 1D};
		
	private double mMomentum = smParameterDefaultValues[0];
	private double mAlphaPlus = smParameterDefaultValues[1];
	private double mAlphaMinus = smParameterDefaultValues[2];
	private double mAlphaInit = smParameterDefaultValues[3];
	
	/* (non-Javadoc)
	 * @see mscJNeuralNet.trainingAlgorithms.INetTrainingAlgorithm#init()
	 */
	public void init(){
		super.init();
		
		// init gradients
		mGradientWOld = new double [getNet().getAmountNeurons()][getNet().getAmountNeurons()];
		mAlphaIJ = new double [getNet().getAmountNeurons()][getNet().getAmountNeurons()];
		mWeightChange = new double [getNet().getAmountNeurons()][getNet().getAmountNeurons()];
		mWeightChangeOld = new double [getNet().getAmountNeurons()][getNet().getAmountNeurons()];
		for (int x=0; x<getNet().getAmountNeurons(); x++){
			for (int y=0; y<getNet().getAmountNeurons(); y++){
				// mGradientW[x][y] = 0D; // wird extern berechnet
				mGradientWOld[x][y] = 0D;
				mAlphaIJ[x][y] = mAlphaInit;
				mWeightChange[x][y] = 0D;
				mWeightChangeOld[x][y] = 0D;
			}
		}
		setCycle(0);	
	}
	
	private double mGradientDirection;
	
	/* (non-Javadoc)
	 * @see mscJNeuralNet.trainingAlgorithms.INetTrainingAlgorithm#calculateWeightUpdate(int, int)
	 */
	public void applyWeightUpdate(int pOutputting, int pInputting){
		super.applyWeightUpdate(pOutputting, pInputting);
		
		mGradientDirection = getActualGradients()[pOutputting][pInputting] * 
			mGradientWOld[pOutputting][pInputting];
		if (mGradientDirection > 0)
			mAlphaIJ[pOutputting][pInputting] *= mAlphaPlus;
		else if (mGradientDirection < 0)
			mAlphaIJ[pOutputting][pInputting] *= mAlphaMinus;
		
		
		mWeightChange[pOutputting][pInputting] =
			-1D*mAlphaIJ[pOutputting][pInputting]*getActualGradients()[pOutputting][pInputting]+
			mMomentum*mWeightChangeOld[pOutputting][pInputting];
		
		getNet().getWeights()[pOutputting][pInputting] += mWeightChange[pOutputting][pInputting];
		mWeightChangeOld[pOutputting][pInputting] = mWeightChange[pOutputting][pInputting];
		mGradientWOld[pOutputting][pInputting] = getActualGradients()[pOutputting][pInputting];
	}
	
	/**
	 * Dieses Lernverfahren untersttzt folgende Parameter: <br/>
	 * <pre>
	 * Index  Bezeichnung   Funktion
	 *   0    Momentum      Momentum-Faktor fr den Momentum-Term
	 *   1    Alpha Plus    Faktor zum Erhhen der variablen Lernschrittweite
	 *   2    Alpha Minus   Faktor zum Verkleinern der variablen Lernschrittweite
	 *   3    Initial Alpha Initialwert fr alle variablen Lernschrittweiten
	 * </pre>
	 * Am Ende der Methode werden die Observer benachrichtigt.
	 *  
	 * @see INetTrainingAlgorithm#setLearningParameters(double[])
	 */
	public void setLearningParameters(double[] pLearningParameters) {
		if (smLogMethodCall) Log.getInstance().writeln(this.getClass().getName()+".setLearningParameters("+Formats.format(pLearningParameters)+")");
		
		if (pLearningParameters.length>0){
			if (pLearningParameters[0] >= 0D) mMomentum = pLearningParameters[0];
			if (pLearningParameters.length>1){
				if (pLearningParameters[1] >= 0D) mAlphaPlus = pLearningParameters[1];
				if (pLearningParameters.length>2){
					if (pLearningParameters[2] >= 0D) mAlphaMinus = pLearningParameters[2];
					if (pLearningParameters.length>3){
						if (pLearningParameters[3] >= 0D) mAlphaInit = pLearningParameters[3];
					}
				}
			}
		}
		getObserverManager().notifyObservers(this);
	}
	
	/**
	 * Dieses Lernverfahren untersttzt folgende Parameter: <br/>
	 * <pre>
	 * Index  Bezeichnung   Funktion
	 *   0    Momentum      Momentum-Faktor fr den Momentum-Term
	 *   1    Alpha Plus    Faktor zum Erhhen der variablen Lernschrittweite
	 *   2    Alpha Minus   Faktor zum Verkleinern der variablen Lernschrittweite
	 *   3    Initial Alpha Initialwert fr alle variablen Lernschrittweiten
	 * </pre>
	 * 
	 * @see mscJNeuralNet.trainingAlgorithms.INetTrainingAlgorithm#getLearningParameters()
	 */
	public double[] getLearningParameters() {
		double [] lParameters = new double[smParameterNames.length];
		lParameters[0] = mMomentum;
		lParameters[1] = mAlphaPlus;
		lParameters[2] = mAlphaMinus;
		lParameters[3] = mAlphaInit;
		return lParameters;
	}

	/**
	 * Dieses Lernverfahren untersttzt folgende Parameter: <br/>
	 * <pre>
	 * Index  Bezeichnung   Funktion
	 *   0    Momentum      Momentum-Faktor fr den Momentum-Term
	 *   1    Alpha Plus    Faktor zum Erhhen der variablen Lernschrittweite
	 *   2    Alpha Minus   Faktor zum Verkleinern der variablen Lernschrittweite
	 *   3    Initial Alpha Initialwert fr alle variablen Lernschrittweiten
	 * </pre>
	 * 
	 * @see mscJNeuralNet.trainingAlgorithms.INetTrainingAlgorithm#getLearningParameterNames()
	 */
	public String[] getLearningParameterNames() {
		return smParameterNames;
	}
	
	/**
	 * Dieses Lernverfahren untersttzt folgende Parameter: <br/>
	 * <pre>
	 * Index  Bezeichnung   Funktion
	 *   0    Momentum      Momentum-Faktor fr den Momentum-Term
	 *   1    Alpha Plus    Faktor zum Erhhen der variablen Lernschrittweite
	 *   2    Alpha Minus   Faktor zum Verkleinern der variablen Lernschrittweite
	 *   3    Initial Alpha Initialwert fr alle variablen Lernschrittweiten
	 * </pre>
	 * 
	 * @see mscJNeuralNet.trainingAlgorithms.INetTrainingAlgorithm#getDefaultLearningParameters()
	 */
	public double [] getDefaultLearningParameters(){
		return smParameterDefaultValues;
	}
	
	/**
	 * Liefert den String "SuperSAB + Momentum" zurck.
	 * 
	 * @see mscJNeuralNet.trainingAlgorithms.INetTrainingAlgorithm#getAlgorithmName()
	 */
	public String getAlgorithmName(){
		return "SuperSAB + Momentum";
	}
	
	public String toString(){
		return "SuperSAB momentum: "+mMomentum+" a+: "+mAlphaPlus+" a-: "+mAlphaMinus+" a0: "+mAlphaInit;
	}
}
